# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import logging
import sys
from urllib.parse import urldefrag

from beartype.typing import Tuple
from rdflib import RDF, RDFS, Graph, Namespace
from rdflib.exceptions import UniquenessError
from rdflib.term import URIRef

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.datetime_conversions import datetime_from_str
from spdx_tools.spdx.model import CreationInfo, ExternalDocumentRef, Version
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.checksum_parser import parse_checksum
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import (
    get_correctly_typed_triples,
    parse_literal,
    parse_spdx_id,
    remove_prefix,
)
from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, SPDX_NAMESPACE


def parse_creation_info(graph: Graph) -> Tuple[CreationInfo, URIRef]:
    logger = Logger()
    namespace, spdx_id, doc_node = parse_namespace_and_spdx_id(graph)
    spec_version = parse_literal(logger, graph, doc_node, SPDX_NAMESPACE.specVersion)
    data_license = parse_literal(
        logger,
        graph,
        doc_node,
        SPDX_NAMESPACE.dataLicense,
        parsing_method=lambda x: remove_prefix(x, LICENSE_NAMESPACE),
    )
    comment = parse_literal(logger, graph, doc_node, RDFS.comment)
    name = parse_literal(logger, graph, doc_node, SPDX_NAMESPACE.name)

    creation_info_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.CreationInfo)
    if not creation_info_node:
        logger.append("CreationInfo does not exist.")
        raise SPDXParsingError([f"Error while parsing document {name}: {logger.get_messages()}"])

    created = parse_literal(
        logger, graph, creation_info_node, SPDX_NAMESPACE.created, parsing_method=datetime_from_str
    )
    license_list_version = parse_literal(
        logger, graph, creation_info_node, SPDX_NAMESPACE.licenseListVersion, parsing_method=Version.from_string
    )
    creator_comment = parse_literal(logger, graph, creation_info_node, RDFS.comment)
    creators = []
    for _, _, creator_literal in get_correctly_typed_triples(
        logger, graph, creation_info_node, SPDX_NAMESPACE.creator
    ):
        creators.append(ActorParser.parse_actor(creator_literal.toPython()))
    if not creators:
        logger.append("No creators provided.")
    external_document_refs = []
    for _, _, external_document_node in get_correctly_typed_triples(
        logger, graph, doc_node, SPDX_NAMESPACE.externalDocumentRef
    ):
        external_document_refs.append(parse_external_document_refs(external_document_node, graph, namespace))

    raise_parsing_error_if_logger_has_messages(logger, "CreationInfo")
    creation_info = construct_or_raise_parsing_error(
        CreationInfo,
        dict(
            spdx_id=spdx_id,
            document_namespace=namespace,
            spdx_version=spec_version,
            name=name,
            data_license=data_license,
            document_comment=comment,
            created=created,
            license_list_version=license_list_version,
            creator_comment=creator_comment,
            creators=creators,
            external_document_refs=external_document_refs,
        ),
    )
    return creation_info, doc_node


def parse_namespace_and_spdx_id(graph: Graph) -> (str, str):
    try:
        subject = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.SpdxDocument, any=False)
    except UniquenessError:
        logging.error("Multiple SpdxDocuments found, can't parse rdf file.")
        sys.exit(1)

    if not subject:
        logging.error("No SpdxDocument found, can't parse rdf file.")
        sys.exit(1)
    if "#" not in subject:
        logging.error(
            "No '#' found in the URI of SpdxDocument, "
            f"the URI for the SpdxDocument should be the namespace appended by '#{DOCUMENT_SPDX_ID}."
        )
        sys.exit(1)

    namespace, spdx_id = urldefrag(str(subject))

    if not namespace:
        logging.error(
            f"No namespace found, the URI for the SpdxDocument should be the namespace appended by "
            f"'#{DOCUMENT_SPDX_ID}."
        )
        sys.exit(1)

    if not spdx_id:
        spdx_id = None

    return namespace, spdx_id, subject


def parse_external_document_refs(
    external_document_node: URIRef, graph: Graph, doc_namespace: str
) -> ExternalDocumentRef:
    logger = Logger()
    document_ref_id = parse_spdx_id(external_document_node, doc_namespace, graph)
    document_uri = parse_literal(logger, graph, external_document_node, SPDX_NAMESPACE.spdxDocument)
    checksum = parse_literal(
        logger,
        graph,
        external_document_node,
        SPDX_NAMESPACE.checksum,
        parsing_method=lambda x: parse_checksum(x, graph),
    )
    external_document_ref = construct_or_raise_parsing_error(
        ExternalDocumentRef, dict(document_ref_id=document_ref_id, document_uri=document_uri, checksum=checksum)
    )

    # To replace the external doc namespaces by the ref id in spdx ids later (e.g. in a relationship), we need to bind
    # the namespace to the graph.
    graph.bind(external_document_ref.document_ref_id, Namespace(external_document_ref.document_uri + "#"))

    return external_document_ref



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

SPDX_NONE_STRING = "NONE"


class SpdxNone:
    """
    Represents the SPDX NONE value.
    """

    def __str__(self):
        return SPDX_NONE_STRING

    def __repr__(self):
        return SPDX_NONE_STRING

    def __eq__(self, other):
        return isinstance(other, SpdxNone)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from spdx_tools.spdx3.bump_from_spdx2.file import bump_file
from spdx_tools.spdx3.model import Hash, HashAlgorithm
from spdx_tools.spdx3.model.software import File
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model.file import File as Spdx2_File
from tests.spdx.fixtures import file_fixture


def test_bump_file():
    payload = Payload()
    document_namespace = "https://doc.namespace"
    spdx2_file: Spdx2_File = file_fixture()
    integrity_method: Hash = Hash(HashAlgorithm.SHA1, "71c4025dd9897b364f3ebbb42c484ff43d00791c")
    expected_new_file_id = f"{document_namespace}#{spdx2_file.spdx_id}"

    bump_file(spdx2_file, payload, document_namespace, [], [])
    file = payload.get_element(expected_new_file_id)

    assert isinstance(file, File)
    assert file.spdx_id == expected_new_file_id
    assert file.verified_using == [integrity_method]
    assert file.copyright_text == spdx2_file.copyright_text
    assert file.attribution_text == spdx2_file.attribution_texts[0]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict
from rdflib import RDF, RDFS, BNode, Graph, Literal, URIRef

from spdx_tools.spdx.casing_tools import snake_case_to_camel_case
from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.model import Annotation
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.writer_utils import add_namespace_to_spdx_id


def add_annotation_to_graph(
    annotation: Annotation, graph: Graph, doc_namespace: str, external_doc_ref_to_namespace: Dict[str, str]
):
    annotation_resource = URIRef(
        add_namespace_to_spdx_id(annotation.spdx_id, doc_namespace, external_doc_ref_to_namespace)
    )
    annotation_node = BNode()
    graph.add((annotation_node, RDF.type, SPDX_NAMESPACE.Annotation))
    graph.add(
        (
            annotation_node,
            SPDX_NAMESPACE.annotationType,
            SPDX_NAMESPACE[f"annotationType_{snake_case_to_camel_case(annotation.annotation_type.name)}"],
        )
    )
    graph.add((annotation_node, SPDX_NAMESPACE.annotator, Literal(annotation.annotator.to_serialized_string())))
    graph.add(
        (annotation_node, SPDX_NAMESPACE.annotationDate, Literal(datetime_to_iso_string(annotation.annotation_date)))
    )
    graph.add((annotation_node, RDFS.comment, Literal(annotation.annotation_comment)))

    graph.add((annotation_resource, SPDX_NAMESPACE.annotation, annotation_node))



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.common.typing.type_checks import check_types_and_set_values
from spdx_tools.spdx3.model import (
    CreationInfo,
    ExternalIdentifier,
    ExternalReference,
    IntegrityMethod,
    RelationshipCompleteness,
)
from spdx_tools.spdx3.model.security.vuln_assessment_relationship import VulnAssessmentRelationship
from spdx_tools.spdx.model import RelationshipType


@dataclass_with_properties
class CvssV3VulnAssessmentRelationship(VulnAssessmentRelationship):
    score: str = None
    severity: Optional[str] = None
    vector: Optional[str] = None

    def __init__(
        self,
        spdx_id: str,
        from_element: str,
        to: List[str],
        relationship_type: RelationshipType,
        score: str,
        creation_info: Optional[CreationInfo] = None,
        name: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        comment: Optional[str] = None,
        verified_using: List[IntegrityMethod] = None,
        external_reference: List[ExternalReference] = None,
        external_identifier: List[ExternalIdentifier] = None,
        extension: Optional[str] = None,
        completeness: Optional[RelationshipCompleteness] = None,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        assessed_element: Optional[str] = None,
        published_time: Optional[datetime] = None,
        supplied_by: Optional[str] = None,
        modified_time: Optional[datetime] = None,
        withdrawn_time: Optional[datetime] = None,
        severity: Optional[str] = None,
        vector: Optional[str] = None,
    ):
        verified_using = [] if verified_using is None else verified_using
        external_reference = [] if external_reference is None else external_reference
        external_identifier = [] if external_identifier is None else external_identifier
        check_types_and_set_values(self, locals())



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List, Optional

from spdx_tools.spdx.model import ChecksumAlgorithm, Document, File
from spdx_tools.spdx.validation.checksum_validator import validate_checksums
from spdx_tools.spdx.validation.license_expression_validator import (
    validate_license_expression,
    validate_license_expressions,
)
from spdx_tools.spdx.validation.spdx_id_validators import validate_spdx_id
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_files(
    files: List[File], spdx_version: str, document: Optional[Document] = None
) -> List[ValidationMessage]:
    validation_messages = []
    if document:
        for file in files:
            validation_messages.extend(validate_file_within_document(file, spdx_version, document))
    else:
        for file in files:
            validation_messages.extend(validate_file(file, spdx_version))

    return validation_messages


def validate_file_within_document(file: File, spdx_version: str, document: Document) -> List[ValidationMessage]:
    validation_messages: List[ValidationMessage] = []
    context = ValidationContext(
        spdx_id=file.spdx_id,
        parent_id=document.creation_info.spdx_id,
        element_type=SpdxElementType.FILE,
        full_element=file,
    )

    for message in validate_spdx_id(file.spdx_id, document):
        validation_messages.append(ValidationMessage(message, context))

    validation_messages.extend(validate_license_expression(file.license_concluded, document, file.spdx_id))

    validation_messages.extend(validate_license_expressions(file.license_info_in_file, document, file.spdx_id))

    validation_messages.extend(validate_file(file, spdx_version, context))

    return validation_messages


def validate_file(
    file: File, spdx_version: str, context: Optional[ValidationContext] = None
) -> List[ValidationMessage]:
    validation_messages = []
    if not context:
        context = ValidationContext(spdx_id=file.spdx_id, element_type=SpdxElementType.FILE, full_element=file)

    if file.name.startswith("/"):
        validation_messages.append(
            ValidationMessage(
                f'file name must not be an absolute path starting with "/", but is: {file.name}', context
            )
        )

    if ChecksumAlgorithm.SHA1 not in [checksum.algorithm for checksum in file.checksums]:
        validation_messages.append(
            ValidationMessage(
                f"checksums must contain a SHA1 algorithm checksum, but only contains: "
                f"{[checksum.algorithm for checksum in file.checksums]}",
                context,
            )
        )

    validation_messages.extend(validate_checksums(file.checksums, file.spdx_id, spdx_version))

    if spdx_version == "SPDX-2.2":
        if file.license_concluded is None:
            validation_messages.append(ValidationMessage("license_concluded is mandatory in SPDX-2.2", context))
        if not file.license_info_in_file:
            validation_messages.append(ValidationMessage("license_info_in_file is mandatory in SPDX-2.2", context))
        if file.copyright_text is None:
            validation_messages.append(ValidationMessage("copyright_text is mandatory in SPDX-2.2", context))

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock, call, mock_open, patch

from spdx_tools.spdx.writer.tagvalue.file_writer import write_file
from tests.spdx.fixtures import file_fixture


def test_file_writer():
    spdx_file = file_fixture()

    mock: MagicMock = mock_open()
    with patch(f"{__name__}.open", mock, create=True):
        with open("foo", "w") as file:
            write_file(spdx_file, file)

    mock.assert_called_once_with("foo", "w")
    handle = mock()
    handle.write.assert_has_calls(
        [
            call("## File Information\n"),
            call(f"FileName: {spdx_file.name}\n"),
            call(f"SPDXID: {spdx_file.spdx_id}\n"),
            call(f"FileType: {spdx_file.file_types[0].name}\n"),
            call("FileChecksum: SHA1: 71c4025dd9897b364f3ebbb42c484ff43d00791c\n"),
            call(f"LicenseConcluded: {spdx_file.license_concluded}\n"),
            call(f"LicenseInfoInFile: {spdx_file.license_info_in_file[0]}\n"),
            call(f"LicenseInfoInFile: {spdx_file.license_info_in_file[1]}\n"),
            call(f"LicenseInfoInFile: {spdx_file.license_info_in_file[2]}\n"),
            call(f"LicenseComments: {spdx_file.license_comment}\n"),
            call(f"FileCopyrightText: {spdx_file.copyright_text}\n"),
            call(f"FileComment: {spdx_file.comment}\n"),
            call(f"FileNotice: {spdx_file.notice}\n"),
            call(f"FileContributor: {spdx_file.contributors[0]}\n"),
            call(f"FileAttributionText: {spdx_file.attribution_texts[0]}\n"),
        ]
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from unittest import mock

import pytest

from spdx_tools.spdx.model import Document


@mock.patch("spdx_tools.spdx.model.ExtractedLicensingInfo", autospec=True)
@mock.patch("spdx_tools.spdx.model.Relationship", autospec=True)
@mock.patch("spdx_tools.spdx.model.Annotation", autospec=True)
@mock.patch("spdx_tools.spdx.model.Snippet", autospec=True)
@mock.patch("spdx_tools.spdx.model.File", autospec=True)
@mock.patch("spdx_tools.spdx.model.Package", autospec=True)
@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_correct_initialization(creation_info, package, file, snippet, annotation, relationship, extracted_lic):
    document = Document(
        creation_info,
        [package, package],
        [file, file],
        [snippet, snippet],
        [annotation, annotation],
        [relationship, relationship],
        [extracted_lic, extracted_lic],
    )
    assert document.creation_info == creation_info
    assert document.packages == [package, package]
    assert document.files == [file, file]
    assert document.snippets == [snippet, snippet]
    assert document.annotations == [annotation, annotation]
    assert document.relationships == [relationship, relationship]
    assert document.extracted_licensing_info == [extracted_lic, extracted_lic]


@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_correct_initialization_with_default_values(creation_info):
    document = Document(creation_info)
    assert document.creation_info == creation_info
    assert document.packages == []
    assert document.files == []
    assert document.snippets == []
    assert document.annotations == []
    assert document.relationships == []
    assert document.extracted_licensing_info == []


def test_wrong_type_in_creation_info():
    with pytest.raises(TypeError):
        Document("string")


@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_wrong_type_in_packages(creation_info):
    with pytest.raises(TypeError):
        Document(creation_info, packages=["string"])


@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_wrong_type_in_files(creation_info):
    with pytest.raises(TypeError):
        Document(creation_info, files={})


@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_wrong_type_in_snippets(creation_info):
    with pytest.raises(TypeError):
        Document(creation_info, snippets=())


@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_wrong_type_in_annotations(creation_info):
    with pytest.raises(TypeError):
        Document(creation_info, annotations=["string"])


@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_wrong_type_in_relationships(creation_info):
    with pytest.raises(TypeError):
        Document(creation_info, relationships="string")


@mock.patch("spdx_tools.spdx.model.CreationInfo", autospec=True)
def test_wrong_type_in_extracted_licensing_info(creation_info):
    with pytest.raises(TypeError):
        Document(creation_info, extracted_licensing_info=42)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Optional, Union
from rdflib import DOAP, RDFS, Graph, URIRef
from rdflib.term import BNode

from spdx_tools.spdx.datetime_conversions import datetime_from_str
from spdx_tools.spdx.model import (
    ExternalPackageRef,
    ExternalPackageRefCategory,
    Package,
    PackagePurpose,
    PackageVerificationCode,
)
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.checksum_parser import parse_checksum
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import (
    get_correctly_typed_triples,
    get_correctly_typed_value,
    get_value_from_graph,
    parse_enum_value,
    parse_literal,
    parse_literal_or_no_assertion_or_none,
    parse_spdx_id,
)
from spdx_tools.spdx.parser.rdf.license_expression_parser import parse_license_expression
from spdx_tools.spdx.rdfschema.namespace import REFERENCE_NAMESPACE, SPDX_NAMESPACE


def parse_package(package_node: Union[URIRef, BNode], graph: Graph, doc_namespace: str) -> Package:
    logger = Logger()
    spdx_id = parse_spdx_id(package_node, doc_namespace, graph)
    name = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.name)
    download_location = parse_literal_or_no_assertion_or_none(
        logger, graph, package_node, SPDX_NAMESPACE.downloadLocation
    )
    checksums = []
    for _, _, checksum_node in get_correctly_typed_triples(logger, graph, package_node, SPDX_NAMESPACE.checksum):
        checksums.append(parse_checksum(checksum_node, graph))

    version_info = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.versionInfo)
    package_file_name = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.packageFileName)

    supplier = parse_literal_or_no_assertion_or_none(
        logger, graph, package_node, SPDX_NAMESPACE.supplier, parsing_method=ActorParser.parse_actor
    )
    originator = parse_literal_or_no_assertion_or_none(
        logger, graph, package_node, SPDX_NAMESPACE.originator, parsing_method=ActorParser.parse_actor
    )
    verification_code = parse_literal(
        logger,
        graph,
        package_node,
        SPDX_NAMESPACE.packageVerificationCode,
        parsing_method=lambda x: parse_package_verification_code(x, graph),
    )

    external_package_refs = []
    for _, _, external_package_ref_node in get_correctly_typed_triples(
        logger, graph, package_node, SPDX_NAMESPACE.externalRef
    ):
        external_package_refs.append(parse_external_package_ref(external_package_ref_node, graph, doc_namespace))
    files_analyzed = bool(
        get_value_from_graph(logger, graph, package_node, SPDX_NAMESPACE.filesAnalyzed, default=True)
    )
    license_concluded = parse_literal_or_no_assertion_or_none(
        logger,
        graph,
        package_node,
        SPDX_NAMESPACE.licenseConcluded,
        parsing_method=lambda x: parse_license_expression(x, graph, doc_namespace, logger),
    )
    license_declared = parse_literal_or_no_assertion_or_none(
        logger,
        graph,
        package_node,
        SPDX_NAMESPACE.licenseDeclared,
        parsing_method=lambda x: parse_license_expression(x, graph, doc_namespace, logger),
    )
    license_info_from_files = []
    for _, _, license_info_from_files_node in graph.triples((package_node, SPDX_NAMESPACE.licenseInfoFromFiles, None)):
        license_info_from_files.append(
            get_correctly_typed_value(
                logger,
                license_info_from_files_node,
                lambda x: parse_license_expression(x, graph, doc_namespace, logger),
            )
        )
    license_comment = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.licenseComments)
    comment = parse_literal(logger, graph, package_node, RDFS.comment)
    summary = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.summary)
    description = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.description)
    copyright_text = parse_literal_or_no_assertion_or_none(logger, graph, package_node, SPDX_NAMESPACE.copyrightText)
    source_info = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.sourceInfo)
    primary_package_purpose = parse_literal(
        logger,
        graph,
        package_node,
        SPDX_NAMESPACE.primaryPackagePurpose,
        parsing_method=lambda x: parse_enum_value(x, PackagePurpose, SPDX_NAMESPACE.purpose_),
    )
    homepage = parse_literal(logger, graph, package_node, DOAP.homepage)
    attribution_texts = []
    for _, _, attribution_text_literal in get_correctly_typed_triples(
        logger, graph, package_node, SPDX_NAMESPACE.attributionText, None
    ):
        attribution_texts.append(attribution_text_literal.toPython())

    release_date = parse_literal(
        logger, graph, package_node, SPDX_NAMESPACE.releaseDate, parsing_method=datetime_from_str
    )
    built_date = parse_literal(logger, graph, package_node, SPDX_NAMESPACE.builtDate, parsing_method=datetime_from_str)
    valid_until_date = parse_literal(
        logger, graph, package_node, SPDX_NAMESPACE.validUntilDate, parsing_method=datetime_from_str
    )
    raise_parsing_error_if_logger_has_messages(logger, "Package")
    package = construct_or_raise_parsing_error(
        Package,
        dict(
            name=name,
            spdx_id=spdx_id,
            download_location=download_location,
            version=version_info,
            file_name=package_file_name,
            supplier=supplier,
            originator=originator,
            files_analyzed=files_analyzed,
            verification_code=verification_code,
            checksums=checksums,
            homepage=homepage,
            source_info=source_info,
            license_concluded=license_concluded,
            license_info_from_files=license_info_from_files,
            license_declared=license_declared,
            license_comment=license_comment,
            copyright_text=copyright_text,
            summary=summary,
            description=description,
            comment=comment,
            external_references=external_package_refs,
            attribution_texts=attribution_texts,
            primary_package_purpose=primary_package_purpose,
            release_date=release_date,
            built_date=built_date,
            valid_until_date=valid_until_date,
        ),
    )

    return package


def parse_package_verification_code(
    package_verification_code_node: URIRef, graph: Graph
) -> Optional[PackageVerificationCode]:
    logger = Logger()
    value = parse_literal(logger, graph, package_verification_code_node, SPDX_NAMESPACE.packageVerificationCodeValue)
    excluded_files = []
    for _, _, excluded_file_literal in graph.triples(
        (package_verification_code_node, SPDX_NAMESPACE.packageVerificationCodeExcludedFile, None)
    ):
        excluded_files.append(excluded_file_literal.toPython())

    raise_parsing_error_if_logger_has_messages(logger, "PackageVerificationCode")
    package_verification_code = construct_or_raise_parsing_error(
        PackageVerificationCode, dict(value=value, excluded_files=excluded_files)
    )
    return package_verification_code


def parse_external_package_ref(external_package_ref_node: BNode, graph: Graph, doc_namespace) -> ExternalPackageRef:
    logger = Logger()
    ref_locator = parse_literal(logger, graph, external_package_ref_node, SPDX_NAMESPACE.referenceLocator)
    ref_category = parse_literal(
        logger,
        graph,
        external_package_ref_node,
        SPDX_NAMESPACE.referenceCategory,
        parsing_method=lambda x: parse_enum_value(x, ExternalPackageRefCategory, SPDX_NAMESPACE.referenceCategory_),
    )
    ref_type = parse_literal(
        logger,
        graph,
        external_package_ref_node,
        SPDX_NAMESPACE.referenceType,
        parsing_method=lambda x: parse_external_package_ref_type(x, doc_namespace),
    )
    comment = parse_literal(logger, graph, external_package_ref_node, RDFS.comment)

    raise_parsing_error_if_logger_has_messages(logger, "ExternalPackageRef")
    external_package_ref = construct_or_raise_parsing_error(
        ExternalPackageRef, dict(category=ref_category, reference_type=ref_type, locator=ref_locator, comment=comment)
    )
    return external_package_ref


def parse_external_package_ref_type(external_package_ref_type_resource: URIRef, doc_namespace: str) -> str:
    if external_package_ref_type_resource.startswith(doc_namespace):
        return external_package_ref_type_resource.fragment
    if external_package_ref_type_resource.startswith(REFERENCE_NAMESPACE):
        return external_package_ref_type_resource.replace(REFERENCE_NAMESPACE, "")
    return external_package_ref_type_resource.toPython()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx.validation.uri_validators import validate_download_location, validate_uri, validate_url


@pytest.mark.parametrize(
    "input_value",
    [
        "https://some.url",
        "https://spdx.org/spdxdocs/spdx-tools-v1.2-3F2504E0-4F89-41D3-9A0C-0305E82...",
        "http://some.url",
        "http://ftp.gnu.org/gnu/glibc/glibc-ports-2.15.tar.gz",
    ],
)
def test_valid_url(input_value):
    assert validate_url(input_value) == []
    # URLs are also valid download locations:
    assert validate_download_location(input_value) == []


# TODO: more negative examples: https://github.com/spdx/tools-python/issues/377
@pytest.mark.parametrize("input_value", [":::::", "http://testurl"])
def test_invalid_url(input_value):
    assert validate_url(input_value) == [f"must be a valid URL, but is: {input_value}"]


@pytest.mark.parametrize(
    "input_value",
    [
        "http://ftp.gnu.org/gnu/glibc/glibc-ports-2.15.tar.gz",
        "git://git.myproject.org/MyProject",
        "git+https://git.myproject.org/MyProject.git",
        "git+http://git.myproject.org/MyProject",
        "git+ssh://git.myproject.org/MyProject.git",
        "git+ssh://git@git.myproject.org/MyProject.git",
        "git+git://git.myproject.org/MyProject",
        "git+git@git.myproject.org:MyProject",
        "git://git.myproject.org/MyProject#src/somefile.c",
        "git+https://git.myproject.org/MyProject#src/Class.java",
        "git://git.myproject.org/MyProject.git@master",
        "git+https://git.myproject.org/MyProject.git@v1.0",
        "git://git.myproject.org/MyProject.git@da39a3ee5e6b4b0d3255bfef95601890afd80709",
        "git+https://git.myproject.org/MyProject.git@master#/src/MyClass.cpp",
        "git+https://git.myproject.org/MyProject@da39a3ee5e6b4b0d3255bfef95601890afd80709#lib/variable.rb",
        "hg+http://hg.myproject.org/MyProject",
        "hg+https://hg.myproject.org/MyProject",
        "hg+ssh://hg.myproject.org/MyProject",
        "hg+https://hg.myproject.org/MyProject#src/somefile.c",
        "hg+https://hg.myproject.org/MyProject#src/Class.java",
        "hg+https://hg.myproject.org/MyProject@da39a3ee5e6b",
        "hg+https://hg.myproject.org/MyProject@2019",
        "hg+https://hg.myproject.org/MyProject@v1.0",
        "hg+https://hg.myproject.org/MyProject@special_feature",
        "hg+https://hg.myproject.org/MyProject@master#/src/MyClass.cpp",
        "hg+https://hg.myproject.org/MyProject@da39a3ee5e6b#lib/variable.rb",
        "svn://svn.myproject.org/svn/MyProject",
        "svn+svn://svn.myproject.org/svn/MyProject",
        "svn+http://svn.myproject.org/svn/MyProject/trunk",
        "svn+https://svn.myproject.org/svn/MyProject/trunk",
        "svn+https://svn.myproject.org/MyProject#src/somefile.c",
        "svn+https://svn.myproject.org/MyProject#src/Class.java",
        "svn+https://svn.myproject.org/MyProject/trunk#src/somefile.c",
        "svn+https://svn.myproject.org/MyProject/trunk/src/somefile.c",
        "svn+https://svn.myproject.org/svn/MyProject/trunk@2019",
        "svn+https://svn.myproject.org/MyProject@123#/src/MyClass.cpp",
        "svn+https://svn.myproject.org/MyProject/trunk@1234#lib/variable/variable.rb",
        "bzr+https://bzr.myproject.org/MyProject/trunk",
        "bzr+http://bzr.myproject.org/MyProject/trunk",
        "bzr+sftp://myproject.org/MyProject/trunk",
        "bzr+ssh://myproject.org/MyProject/trunk",
        "bzr+ftp://myproject.org/MyProject/trunk",
        "bzr+lp:MyProject",
        "bzr+https://bzr.myproject.org/MyProject/trunk#src/somefile.c",
        "bzr+https://bzr.myproject.org/MyProject/trunk#src/Class.java",
        "bzr+https://bzr.myproject.org/MyProject/trunk@2019",
        "bzr+http://bzr.myproject.org/MyProject/trunk@v1.0",
        "bzr+https://bzr.myproject.org/MyProject/trunk@2019#src/somefile.c",
    ],
)
def test_valid_package_download_location(input_value):
    assert validate_download_location(input_value) == []


# TODO: more negative examples: https://github.com/spdx/tools-python/issues/377
@pytest.mark.parametrize(
    "input_value",
    [
        ":::::",
    ],
)
def test_invalid_package_download_location(input_value):
    assert validate_download_location(input_value) == [
        f"must be a valid URL or download location according to the specification, but is: {input_value}"
    ]


@pytest.mark.parametrize(
    "input_value",
    [
        "https://some.uri",
        "http:////some",
        "https://spdx.org/spdxdocs/spdx-tools-v1.2-3F2504E0-4F89-41D3-9A0C-0305E82...",
        "h://someweirdtest^?",
        "https://some.uri that goes on!?",
    ],
)
def test_valid_uri(input_value):
    message = validate_uri(input_value)

    assert message == []


@pytest.mark.parametrize(
    "input_value", ["/invalid/uri", "http//uri", "http://some#uri", "some/uri", "some weird test"]
)
def test_invalid_uri(input_value):
    message = validate_uri(input_value)

    assert message == [
        f"must be a valid URI specified in RFC-3986 and must contain no fragment (#), but is: {input_value}"
    ]


@pytest.mark.parametrize("input_value", ["://spdx.org/spdxdocs/spdx-tools-v1.2-3F2504E0-4F89-41D3-9A0C-0305E82..."])
@pytest.mark.skip(
    "validate_uri() seems to invalidate URIs without scheme, so it does not run into this case. But I'm not sure yet "
    "if this covers all scheme-less examples."
    "https://github.com/spdx/tools-python/issues/377"
)
def test_uri_without_scheme(input_value):
    message = validate_uri(input_value)

    assert message == [f"must have a URI scheme, but is: {input_value}"]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum, auto


class SoftwarePurpose(Enum):
    APPLICATION = auto()
    ARCHIVE = auto()
    BOM = auto()
    CONFIGURATION = auto()
    CONTAINER = auto()
    DATA = auto()
    DEVICE = auto()
    DOCUMENTATION = auto()
    EXECUTABLE = auto()
    FILE = auto()
    FIRMWARE = auto()
    FRAMEWORK = auto()
    INSTALL = auto()
    LIBRARY = auto()
    MODEL = auto()
    MODULE = auto()
    OPERATING_SYSTEM = auto()
    OTHER = auto()
    PATCH = auto()
    REQUIREMENT = auto()
    SOURCE = auto()
    TEST = auto()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from datetime import datetime
from unittest import mock

import pytest

from spdx_tools.spdx.model import CreationInfo, Version


@mock.patch("spdx_tools.spdx.model.ExternalDocumentRef", autospec=True)
@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_correct_initialization(actor, ext_ref):
    creation_info = CreationInfo(
        "version",
        "id",
        "name",
        "namespace",
        [actor, actor],
        datetime(2022, 1, 1),
        "creator_comment",
        "CC0-1.1",
        [ext_ref, ext_ref],
        Version(6, 3),
        "doc_comment",
    )
    assert creation_info.spdx_version == "version"
    assert creation_info.spdx_id == "id"
    assert creation_info.name == "name"
    assert creation_info.document_namespace == "namespace"
    assert creation_info.creators == [actor, actor]
    assert creation_info.created == datetime(2022, 1, 1)
    assert creation_info.creator_comment == "creator_comment"
    assert creation_info.data_license == "CC0-1.1"
    assert creation_info.external_document_refs == [ext_ref, ext_ref]
    assert creation_info.license_list_version == Version(6, 3)
    assert creation_info.document_comment == "doc_comment"


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_spdx_version(actor):
    with pytest.raises(TypeError):
        CreationInfo(42, "id", "name", "namespace", [actor, actor], datetime(2022, 1, 1))


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_spdx_id(actor):
    with pytest.raises(TypeError):
        CreationInfo("version", 42, "name", "namespace", [actor, actor], datetime(2022, 1, 1))


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_name(actor):
    with pytest.raises(TypeError):
        CreationInfo("version", "id", 42, "namespace", [actor, actor], datetime(2022, 1, 1))


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_document_namespace(actor):
    with pytest.raises(TypeError):
        CreationInfo("version", "id", "name", 42, [actor, actor], datetime(2022, 1, 1))


def test_wrong_type_in_creators():
    with pytest.raises(TypeError):
        CreationInfo("version", "id", "name", "namespace", ["person"], datetime(2022, 1, 1))


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_created(actor):
    with pytest.raises(TypeError):
        CreationInfo("version", "id", "name", "namespace", [actor, actor], "2022-01-01")


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_creator_comment(actor):
    with pytest.raises(TypeError):
        CreationInfo(
            "version", "id", "name", "namespace", [actor, actor], datetime(2022, 1, 1), creator_comment=["string"]
        )


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_data_license(actor):
    with pytest.raises(TypeError):
        CreationInfo("version", "id", "name", "namespace", [actor, actor], datetime(2022, 1, 1), data_license=42)


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_external_document_refs(actor):
    with pytest.raises(TypeError):
        CreationInfo(
            "version", "id", "name", "namespace", [actor, actor], datetime(2022, 1, 1), external_document_refs=()
        )


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_license_list_version(actor):
    with pytest.raises(TypeError):
        CreationInfo(
            "version", "id", "name", "namespace", [actor, actor], datetime(2022, 1, 1), license_list_version="6.4"
        )


@mock.patch("spdx_tools.spdx.model.Actor", autospec=True)
def test_wrong_type_in_document_comment(actor):
    with pytest.raises(TypeError):
        CreationInfo(
            "version", "id", "name", "namespace", [actor, actor], datetime(2022, 1, 1), document_comment=["1"]
        )



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import hashlib

from beartype.typing import List, Optional, Type, Union

from spdx_tools.spdx.model import (
    ChecksumAlgorithm,
    Document,
    ExternalDocumentRef,
    File,
    Package,
    PackageVerificationCode,
    Snippet,
)


def get_element_type_from_spdx_id(
    spdx_id: str, document: Document
) -> Optional[Union[Type[Package], Type[File], Type[Snippet]]]:
    if spdx_id in [package.spdx_id for package in document.packages]:
        return Package
    if spdx_id in [file.spdx_id for file in document.files]:
        return File
    if spdx_id in [snippet.spdx_id for snippet in document.snippets]:
        return Snippet
    return None


def get_full_element_spdx_id(
    element: Union[Package, File, Snippet],
    document_namespace: str,
    external_document_refs: List[ExternalDocumentRef],
) -> str:
    """
    Returns the spdx_id of the element prefixed with the correct document namespace and,
    if the element is from an external document, sets the correct entry in the imports property.
    """
    if ":" not in element.spdx_id:
        return f"{document_namespace}#{element.spdx_id}"

    external_id, local_id = element.spdx_id.split(":")
    external_uri = None
    for entry in external_document_refs:
        if entry.document_ref_id == external_id:
            external_uri = entry.document_uri
            break

    if not external_uri:
        raise ValueError(f"external id {external_id} not found in external document references")

    return external_uri + "#" + local_id


def calculate_package_verification_code(files: List[File]) -> PackageVerificationCode:
    list_of_file_hashes = []
    for file in files:
        file_checksum_value = None
        for checksum in file.checksums:
            if checksum.algorithm == ChecksumAlgorithm.SHA1:
                file_checksum_value = checksum.value
        if not file_checksum_value:
            try:
                file_checksum_value = calculate_file_checksum(file.name, ChecksumAlgorithm.SHA1)
            except FileNotFoundError:
                raise FileNotFoundError(
                    f"Cannot calculate package verification code because the file '{file.name}' "
                    f"provides no SHA1 checksum and can't be found at the specified location."
                )
        list_of_file_hashes.append(file_checksum_value)

    list_of_file_hashes.sort()
    hasher = hashlib.new("sha1")
    hasher.update("".join(list_of_file_hashes).encode("utf-8"))
    value = hasher.hexdigest()
    return PackageVerificationCode(value)


def calculate_file_checksum(file_name: str, hash_algorithm=ChecksumAlgorithm.SHA1) -> str:
    BUFFER_SIZE = 65536

    file_hash = hashlib.new(hash_algorithm.name.lower())
    with open(file_name, "rb") as file_handle:
        while True:
            data = file_handle.read(BUFFER_SIZE)
            if not data:
                break
            file_hash.update(data)

    return file_hash.hexdigest()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os
from datetime import datetime
from typing import List, Tuple

import pytest
from rdflib import RDF, Graph, URIRef
from rdflib.term import Node

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Actor, ActorType, Checksum, ChecksumAlgorithm, Version
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.rdf.creation_info_parser import (
    parse_creation_info,
    parse_external_document_refs,
    parse_namespace_and_spdx_id,
)
from spdx_tools.spdx.parser.rdf.rdf_parser import parse_from_file
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def test_parse_creation_info():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))

    creation_info, _ = parse_creation_info(graph)
    assert creation_info.spdx_id == DOCUMENT_SPDX_ID
    assert creation_info.spdx_version == "SPDX-2.3"
    assert creation_info.name == "documentName"
    assert creation_info.document_namespace == "https://some.namespace"
    assert creation_info.creators == [Actor(ActorType.PERSON, "creatorName", "some@mail.com")]
    assert creation_info.created == datetime(2022, 12, 1, 0, 0)
    assert creation_info.creator_comment == "creatorComment"
    assert creation_info.data_license == "CC0-1.0"
    assert creation_info.license_list_version == Version(3, 19)
    assert creation_info.document_comment == "documentComment"


def test_parse_namespace_and_spdx_id():
    graph = Graph().add((URIRef("docNamespace#spdxID"), RDF.type, SPDX_NAMESPACE.SpdxDocument))

    namespace, spdx_id, _ = parse_namespace_and_spdx_id(graph)

    assert namespace == "docNamespace"
    assert spdx_id == "spdxID"


@pytest.mark.parametrize(
    "triples,error_message",
    [
        (
            [(URIRef("docNamespace"), RDF.type, SPDX_NAMESPACE.SpdxDocument)],
            r"No '#' found in the URI of SpdxDocument",
        ),
        ([(URIRef(""), RDF.type, URIRef(""))], r"No SpdxDocument found, can't parse rdf file."),
        ([(URIRef(f"#{DOCUMENT_SPDX_ID}"), RDF.type, SPDX_NAMESPACE.SpdxDocument)], "No namespace found"),
        (
            [
                (URIRef("docNamespace1"), RDF.type, SPDX_NAMESPACE.SpdxDocument),
                (URIRef("docNamespace2"), RDF.type, SPDX_NAMESPACE.SpdxDocument),
            ],
            "Multiple SpdxDocuments found",
        ),
    ],
)
def test_parse_namespace_and_spdx_id_with_system_exit(
    triples: List[Tuple[Node, Node, Node]], error_message: str, caplog
):
    graph = Graph()
    for triple in triples:
        graph = graph.add(triple)

    with pytest.raises(SystemExit):
        parse_namespace_and_spdx_id(graph)

    assert error_message in caplog.text


def test_parse_external_document_refs():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    doc_namespace = "https://some.namespace"
    external_doc_ref_node = graph.value(
        subject=URIRef(f"{doc_namespace}#{DOCUMENT_SPDX_ID}"), predicate=SPDX_NAMESPACE.externalDocumentRef
    )
    assert isinstance(external_doc_ref_node, URIRef)

    external_document_ref = parse_external_document_refs(external_doc_ref_node, graph, doc_namespace)

    assert external_document_ref.document_ref_id == "DocumentRef-external"
    assert external_document_ref.checksum == Checksum(
        ChecksumAlgorithm.SHA1, "71c4025dd9897b364f3ebbb42c484ff43d00791c"
    )
    assert external_document_ref.document_uri == "https://namespace.com"


@pytest.mark.parametrize(
    "file, error_message",
    [
        (
            "invalid_creation_info.rdf.xml",
            "Error while parsing CreationInfo: ['No creators provided.']",
        ),
        ("invalid_creation_info_with_snippet.rdf.xml", "Error while parsing CreationInfo: ['No creators provided.']"),
    ],
)
def test_parse_invalid_creation_info(file, error_message):
    with pytest.raises(SPDXParsingError) as err:
        parse_from_file(os.path.join(os.path.dirname(__file__), f"data/invalid_documents/{file}"))

    assert err.value.get_messages() == [error_message]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import Enum

from beartype.typing import Optional, TextIO, Union


def write_value(tag: str, value: Optional[Union[bool, str, dict, list, Enum]], out: TextIO, indent: bool = False):
    """This function is duplicated from spdx_tools.spdx.writer.tagvalue.tag_value_writer_helper_functions
    and slightly adapted to make indentation of output possible."""
    if not value:
        return
    if isinstance(value, dict):
        value = ", ".join([f"{tag}: ({key}: {val})" for key, val in value.items()])
    if isinstance(value, list):
        value = ", ".join([entry for entry in value])
    if isinstance(value, Enum):
        value = value.name
    write_and_possibly_indent(f"{tag}: {value}", indent, out)


def write_and_possibly_indent(text: str, indent: bool, out: TextIO):
    if indent:
        out.write(f"  {text}\n")
    else:
        out.write(f"{text}\n")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import TextIO

from spdx_tools.spdx3.model import LifecycleScopedRelationship
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.relationship_writer import write_relationship


def write_lifecycle_scoped_relationship(
    relationship: LifecycleScopedRelationship, text_output: TextIO, heading: bool = True
):
    if heading:
        text_output.write("## LifecycleScopedRelationship\n")
    write_relationship(relationship, text_output, heading=False)

    for property_name in LifecycleScopedRelationship.__annotations__.keys():
        write_value(property_name, getattr(relationship, property_name), text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from dataclasses import field

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.spdx3.model import Artifact
from spdx_tools.spdx3.model.licensing import LicenseField
from spdx_tools.spdx3.model.software import SoftwarePurpose


@dataclass_with_properties
class SoftwareArtifact(Artifact):
    content_identifier: Optional[str] = None
    primary_purpose: Optional[SoftwarePurpose] = None
    additional_purpose: List[SoftwarePurpose] = field(default_factory=list)
    concluded_license: Optional[LicenseField] = None
    declared_license: Optional[LicenseField] = None
    copyright_text: Optional[str] = None
    attribution_text: Optional[str] = None

    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import TextIO

from spdx_tools.spdx3.model.software import SoftwareDependencyRelationship
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.lifecycle_scoped_relationship_writer import write_lifecycle_scoped_relationship


def write_software_dependency_relationship(
    relationship: SoftwareDependencyRelationship, text_output: TextIO, heading: bool = True
):
    if heading:
        text_output.write("## SoftwareDependencyRelationship\n")
    write_lifecycle_scoped_relationship(relationship, text_output, heading=False)

    for property_name in SoftwareDependencyRelationship.__annotations__.keys():
        write_value(property_name, getattr(relationship, property_name), text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDFS, Graph, Literal, URIRef

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.relationship_writer import add_relationship_to_graph
from tests.spdx.fixtures import relationship_fixture


def test_add_relationship_to_graph():
    relationship = relationship_fixture()
    graph = Graph()
    add_relationship_to_graph(relationship, graph, "docNamespace", {})

    assert (URIRef(f"docNamespace#{DOCUMENT_SPDX_ID}"), SPDX_NAMESPACE.relationship, None) in graph
    assert (None, SPDX_NAMESPACE.relationshipType, SPDX_NAMESPACE.relationshipType_describes) in graph
    assert (None, SPDX_NAMESPACE.relatedSpdxElement, URIRef("docNamespace#SPDXRef-File")) in graph
    assert (None, RDFS.comment, Literal(relationship.comment)) in graph



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List

from spdx_tools.spdx.document_utils import create_document_without_duplicates
from spdx_tools.spdx.jsonschema.document_converter import DocumentConverter
from spdx_tools.spdx.model import Document
from spdx_tools.spdx.validation.document_validator import validate_full_spdx_document
from spdx_tools.spdx.validation.validation_message import ValidationMessage


def validate_and_deduplicate(document: Document, validate: bool = True, drop_duplicates: bool = True) -> Document:
    if validate:
        validation_messages: List[ValidationMessage] = validate_full_spdx_document(document)
        if validation_messages:
            raise ValueError(f"Document is not valid. The following errors were detected: {validation_messages}")
    if drop_duplicates:
        document = create_document_without_duplicates(document)
    return document


def convert(document: Document, converter: DocumentConverter) -> dict:
    if converter is None:
        converter = DocumentConverter()
    return converter.convert(document)



# Copyright (c) 2014 Ahmed H. Ismail
# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.parser.tagvalue.lexer import SPDXLexer


@pytest.fixture
def lexer():
    lexer = SPDXLexer()
    lexer.build()
    return lexer


def token_assert_helper(token, token_type, value, line_number):
    assert token.type == token_type
    assert token.value == value
    assert token.lineno == line_number


def test_tokenization_of_document(lexer):
    document_str = "\n".join(
        [
            "SPDXVersion: SPDX-2.1",
            "DataLicense: CC0-1.0",
            "DocumentName: Sample_Document-V2.1",
            f"SPDXID: {DOCUMENT_SPDX_ID}",
            "DocumentComment: <text>Sample Comment</text>",
            "DocumentNamespace: https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301",
        ]
    )
    lexer.input(document_str)
    token_assert_helper(lexer.token(), "DOC_VERSION", "SPDXVersion", 1)
    token_assert_helper(lexer.token(), "LINE", "SPDX-2.1", 1)
    token_assert_helper(lexer.token(), "DOC_LICENSE", "DataLicense", 2)
    token_assert_helper(lexer.token(), "LINE", "CC0-1.0", 2)
    token_assert_helper(lexer.token(), "DOC_NAME", "DocumentName", 3)
    token_assert_helper(lexer.token(), "LINE", "Sample_Document-V2.1", 3)
    token_assert_helper(lexer.token(), "SPDX_ID", "SPDXID", 4)
    token_assert_helper(lexer.token(), "LINE", DOCUMENT_SPDX_ID, 4)
    token_assert_helper(lexer.token(), "DOC_COMMENT", "DocumentComment", 5)
    token_assert_helper(lexer.token(), "TEXT", "<text>Sample Comment</text>", 5)
    token_assert_helper(lexer.token(), "DOC_NAMESPACE", "DocumentNamespace", 6)
    token_assert_helper(
        lexer.token(), "LINE", "https://spdx.org/spdxdocs/spdx-example-444504E0-4F89-41D3-9A0C-0305E82C3301", 6
    )


def test_tokenization_of_external_document_references(lexer):
    data = "\n".join(
        [
            "ExternalDocumentRef:DocumentRef-spdx-tool-2.1 http://spdx.org/spdxdocs/spdx-tools-v2.1-3F2504E0-4F89-41D3"
            "-9A0C-0305E82C3301 SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
            "ExternalDocumentRef:DocumentRef-spdx-tool-2.1 ldap://[2001:db8::7]/c=GB?objectClass?one "
            "SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
        ]
    )
    lexer.input(data)
    token_assert_helper(lexer.token(), "EXT_DOC_REF", "ExternalDocumentRef", 1)
    token_assert_helper(
        lexer.token(),
        "LINE",
        "DocumentRef-spdx-tool-2.1 http://spdx.org/spdxdocs/spdx-tools-v2.1-3F2504E0-4F89-41D3-9A0C-0305E82C3301 "
        "SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
        1,
    )
    token_assert_helper(lexer.token(), "EXT_DOC_REF", "ExternalDocumentRef", 2)
    token_assert_helper(
        lexer.token(),
        "LINE",
        "DocumentRef-spdx-tool-2.1 ldap://[2001:db8::7]/c=GB?objectClass?one "
        "SHA1: d6a770ba38583ed4bb4525bd96e50461655d2759",
        2,
    )


def test_tokenization_of_file(lexer):
    file_str = "\n".join(
        [
            "FileName: testfile.java",
            "SPDXID: SPDXRef-File",
            "FileType: SOURCE",
            "FileType: TEXT",
            "FileChecksum: SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12",
            "LicenseConcluded: Apache-2.0",
            "LicenseInfoInFile: Apache-2.0",
            "FileCopyrightText: <text>Copyright 2014 Acme Inc.</text>",
            "FileComment: <text>Very long file</text>",
            "FileAttributionText: <text>Acknowledgements that might be required to be communicated in some contexts."
            "</text>",
        ]
    )

    lexer.input(file_str)
    token_assert_helper(lexer.token(), "FILE_NAME", "FileName", 1)
    token_assert_helper(lexer.token(), "LINE", "testfile.java", 1)
    token_assert_helper(lexer.token(), "SPDX_ID", "SPDXID", 2)
    token_assert_helper(lexer.token(), "LINE", "SPDXRef-File", 2)
    token_assert_helper(lexer.token(), "FILE_TYPE", "FileType", 3)
    token_assert_helper(lexer.token(), "LINE", "SOURCE", 3)
    token_assert_helper(lexer.token(), "FILE_TYPE", "FileType", 4)
    token_assert_helper(lexer.token(), "LINE", "TEXT", 4)
    token_assert_helper(lexer.token(), "FILE_CHECKSUM", "FileChecksum", 5)
    token_assert_helper(lexer.token(), "CHECKSUM", "SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12", 5)
    token_assert_helper(lexer.token(), "FILE_LICENSE_CONCLUDED", "LicenseConcluded", 6)
    token_assert_helper(lexer.token(), "LINE", "Apache-2.0", 6)
    token_assert_helper(lexer.token(), "FILE_LICENSE_INFO", "LicenseInfoInFile", 7)
    token_assert_helper(lexer.token(), "LINE", "Apache-2.0", 7)
    token_assert_helper(lexer.token(), "FILE_COPYRIGHT_TEXT", "FileCopyrightText", 8)
    token_assert_helper(lexer.token(), "TEXT", "<text>Copyright 2014 Acme Inc.</text>", 8)
    token_assert_helper(lexer.token(), "FILE_COMMENT", "FileComment", 9)
    token_assert_helper(lexer.token(), "TEXT", "<text>Very long file</text>", 9)
    token_assert_helper(lexer.token(), "FILE_ATTRIBUTION_TEXT", "FileAttributionText", 10)
    token_assert_helper(
        lexer.token(),
        "TEXT",
        "<text>Acknowledgements that might be required to be communicated in some contexts.</text>",
        10,
    )


def test_tokenization_of_creation_info(lexer):
    creation_str = "\n".join(
        [
            "Creator: Person: Bob (bob@example.com)",
            "Creator: Organization: Acme.",
            "Created: 2010-02-03T00:00:00Z",
            "CreatorComment: <text>Sample Comment</text>",
        ]
    )

    lexer.input(creation_str)
    token_assert_helper(lexer.token(), "CREATOR", "Creator", 1)
    token_assert_helper(lexer.token(), "PERSON_VALUE", "Person: Bob (bob@example.com)", 1)
    token_assert_helper(lexer.token(), "CREATOR", "Creator", 2)
    token_assert_helper(lexer.token(), "ORGANIZATION_VALUE", "Organization: Acme.", 2)
    token_assert_helper(lexer.token(), "CREATED", "Created", 3)
    token_assert_helper(lexer.token(), "ISO8601_DATE", "2010-02-03T00:00:00Z", 3)
    token_assert_helper(lexer.token(), "CREATOR_COMMENT", "CreatorComment", 4)
    token_assert_helper(lexer.token(), "TEXT", "<text>Sample Comment</text>", 4)


def test_tokenization_of_package(lexer):
    package_str = "\n".join(
        [
            "PackageName: Test",
            "SPDXID: SPDXRef-Package",
            "PackageVersion: Version 0.9.2",
            "PackageDownloadLocation: http://example.com/test",
            "FilesAnalyzed: True",
            "PackageSummary: <text>Test package</text>",
            "PackageSourceInfo: <text>Version 1.0 of test</text>",
            "PackageFileName: test-1.0.zip",
            "PackageSupplier: Organization:ACME",
            "PackageOriginator: Organization:ACME",
            "PackageChecksum: SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12",
            "PackageVerificationCode: 4e3211c67a2d28fced849ee1bb76e7391b93feba (something.rdf, something.txt)",
            "PackageDescription: <text>A package.</text>",
            "PackageComment: <text>Comment on the package.</text>",
            "PackageCopyrightText: <text> Copyright 2014 Acme Inc.</text>",
            "PackageLicenseDeclared: Apache-2.0",
            "PackageLicenseConcluded: (LicenseRef-2.0 and Apache-2.0)",
            "PackageLicenseInfoFromFiles: Apache-1.0",
            "PackageLicenseInfoFromFiles: Apache-2.0",
            "PackageLicenseComments: <text>License Comments</text>",
            "ExternalRef: SECURITY cpe23Type cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:",
            "ExternalRefComment: <text>Some comment about the package.</text>",
            "PrimaryPackagePurpose: OPERATING-SYSTEM",
            "BuiltDate: 2020-01-01T12:00:00Z",
            "ReleaseDate: 2021-01-01T12:00:00Z",
            "ValidUntilDate: 2022-01-01T12:00:00Z",
        ]
    )

    lexer.input(package_str)
    token_assert_helper(lexer.token(), "PKG_NAME", "PackageName", 1)
    token_assert_helper(lexer.token(), "LINE", "Test", 1)
    token_assert_helper(lexer.token(), "SPDX_ID", "SPDXID", 2)
    token_assert_helper(lexer.token(), "LINE", "SPDXRef-Package", 2)
    token_assert_helper(lexer.token(), "PKG_VERSION", "PackageVersion", 3)
    token_assert_helper(lexer.token(), "LINE", "Version 0.9.2", 3)
    token_assert_helper(lexer.token(), "PKG_DOWNLOAD_LOCATION", "PackageDownloadLocation", 4)
    token_assert_helper(lexer.token(), "LINE", "http://example.com/test", 4)
    token_assert_helper(lexer.token(), "PKG_FILES_ANALYZED", "FilesAnalyzed", 5)
    token_assert_helper(lexer.token(), "LINE", "True", 5)
    token_assert_helper(lexer.token(), "PKG_SUMMARY", "PackageSummary", 6)
    token_assert_helper(lexer.token(), "TEXT", "<text>Test package</text>", 6)
    token_assert_helper(lexer.token(), "PKG_SOURCE_INFO", "PackageSourceInfo", 7)
    token_assert_helper(lexer.token(), "TEXT", "<text>Version 1.0 of test</text>", 7)
    token_assert_helper(lexer.token(), "PKG_FILE_NAME", "PackageFileName", 8)
    token_assert_helper(lexer.token(), "LINE", "test-1.0.zip", 8)
    token_assert_helper(lexer.token(), "PKG_SUPPLIER", "PackageSupplier", 9)
    token_assert_helper(lexer.token(), "ORGANIZATION_VALUE", "Organization:ACME", 9)
    token_assert_helper(lexer.token(), "PKG_ORIGINATOR", "PackageOriginator", 10)
    token_assert_helper(lexer.token(), "ORGANIZATION_VALUE", "Organization:ACME", 10)
    token_assert_helper(lexer.token(), "PKG_CHECKSUM", "PackageChecksum", 11)
    token_assert_helper(lexer.token(), "CHECKSUM", "SHA1: 2fd4e1c67a2d28fced849ee1bb76e7391b93eb12", 11)
    token_assert_helper(lexer.token(), "PKG_VERIFICATION_CODE", "PackageVerificationCode", 12)
    token_assert_helper(
        lexer.token(), "LINE", "4e3211c67a2d28fced849ee1bb76e7391b93feba (something.rdf, something.txt)", 12
    )
    token_assert_helper(lexer.token(), "PKG_DESCRIPTION", "PackageDescription", 13)
    token_assert_helper(lexer.token(), "TEXT", "<text>A package.</text>", 13)
    token_assert_helper(lexer.token(), "PKG_COMMENT", "PackageComment", 14)
    token_assert_helper(lexer.token(), "TEXT", "<text>Comment on the package.</text>", 14)
    token_assert_helper(lexer.token(), "PKG_COPYRIGHT_TEXT", "PackageCopyrightText", 15)
    token_assert_helper(lexer.token(), "TEXT", "<text> Copyright 2014 Acme Inc.</text>", 15)
    token_assert_helper(lexer.token(), "PKG_LICENSE_DECLARED", "PackageLicenseDeclared", 16)
    token_assert_helper(lexer.token(), "LINE", "Apache-2.0", 16)
    token_assert_helper(lexer.token(), "PKG_LICENSE_CONCLUDED", "PackageLicenseConcluded", 17)
    token_assert_helper(lexer.token(), "LINE", "(LicenseRef-2.0 and Apache-2.0)", 17)
    token_assert_helper(lexer.token(), "PKG_LICENSE_INFO", "PackageLicenseInfoFromFiles", 18)
    token_assert_helper(lexer.token(), "LINE", "Apache-1.0", 18)
    token_assert_helper(lexer.token(), "PKG_LICENSE_INFO", "PackageLicenseInfoFromFiles", 19)
    token_assert_helper(lexer.token(), "LINE", "Apache-2.0", 19)
    token_assert_helper(lexer.token(), "PKG_LICENSE_COMMENT", "PackageLicenseComments", 20)
    token_assert_helper(lexer.token(), "TEXT", "<text>License Comments</text>", 20)
    token_assert_helper(lexer.token(), "PKG_EXTERNAL_REF", "ExternalRef", 21)
    token_assert_helper(
        lexer.token(), "LINE", "SECURITY cpe23Type cpe:2.3:a:pivotal_software:spring_framework:4.1.0:*:*:*:*:*:*:", 21
    )
    token_assert_helper(lexer.token(), "PKG_EXTERNAL_REF_COMMENT", "ExternalRefComment", 22)
    token_assert_helper(lexer.token(), "TEXT", "<text>Some comment about the package.</text>", 22)
    token_assert_helper(lexer.token(), "PRIMARY_PACKAGE_PURPOSE", "PrimaryPackagePurpose", 23)
    token_assert_helper(lexer.token(), "LINE", "OPERATING-SYSTEM", 23)
    token_assert_helper(lexer.token(), "BUILT_DATE", "BuiltDate", 24)
    token_assert_helper(lexer.token(), "ISO8601_DATE", "2020-01-01T12:00:00Z", 24)
    token_assert_helper(lexer.token(), "RELEASE_DATE", "ReleaseDate", 25)
    token_assert_helper(lexer.token(), "ISO8601_DATE", "2021-01-01T12:00:00Z", 25)
    token_assert_helper(lexer.token(), "VALID_UNTIL_DATE", "ValidUntilDate", 26)
    token_assert_helper(lexer.token(), "ISO8601_DATE", "2022-01-01T12:00:00Z", 26)


def test_tokenization_of_unknown_tag(lexer):
    unknown_tag_str = "SomeUnknownTag: SomeUnknownValue"
    lexer.input(unknown_tag_str)
    token_assert_helper(lexer.token(), "UNKNOWN_TAG", "SomeUnknownTag", 1)
    token_assert_helper(lexer.token(), "LINE", "SomeUnknownValue", 1)


def test_tokenization_of_snippet(lexer):
    snippet_str = "\n".join(
        [
            "SnippetSPDXID: SPDXRef-Snippet",
            "SnippetLicenseComments: <text>Some lic comment.</text>",
            "SnippetCopyrightText: <text> Copyright 2008-2010 John Smith </text>",
            "SnippetComment: <text>Some snippet comment.</text>",
            "SnippetName: from linux kernel",
            "SnippetFromFileSPDXID: SPDXRef-DoapSource",
            "SnippetLicenseConcluded: Apache-2.0",
            "LicenseInfoInSnippet: Apache-2.0",
            "SnippetByteRange: 310:420",
            "SnippetLineRange: 5:23",
        ]
    )
    lexer.input(snippet_str)
    token_assert_helper(lexer.token(), "SNIPPET_SPDX_ID", "SnippetSPDXID", 1)
    token_assert_helper(lexer.token(), "LINE", "SPDXRef-Snippet", 1)
    token_assert_helper(lexer.token(), "SNIPPET_LICENSE_COMMENT", "SnippetLicenseComments", 2)
    token_assert_helper(lexer.token(), "TEXT", "<text>Some lic comment.</text>", 2)
    token_assert_helper(lexer.token(), "SNIPPET_COPYRIGHT_TEXT", "SnippetCopyrightText", 3)
    token_assert_helper(lexer.token(), "TEXT", "<text> Copyright 2008-2010 John Smith </text>", 3)
    token_assert_helper(lexer.token(), "SNIPPET_COMMENT", "SnippetComment", 4)
    token_assert_helper(lexer.token(), "TEXT", "<text>Some snippet comment.</text>", 4)
    token_assert_helper(lexer.token(), "SNIPPET_NAME", "SnippetName", 5)
    token_assert_helper(lexer.token(), "LINE", "from linux kernel", 5)
    token_assert_helper(lexer.token(), "SNIPPET_FILE_SPDXID", "SnippetFromFileSPDXID", 6)
    token_assert_helper(lexer.token(), "LINE", "SPDXRef-DoapSource", 6)
    token_assert_helper(lexer.token(), "SNIPPET_LICENSE_CONCLUDED", "SnippetLicenseConcluded", 7)
    token_assert_helper(lexer.token(), "LINE", "Apache-2.0", 7)
    token_assert_helper(lexer.token(), "SNIPPET_LICENSE_INFO", "LicenseInfoInSnippet", 8)
    token_assert_helper(lexer.token(), "LINE", "Apache-2.0", 8)
    token_assert_helper(lexer.token(), "SNIPPET_BYTE_RANGE", "SnippetByteRange", 9)
    token_assert_helper(lexer.token(), "LINE", "310:420", 9)
    token_assert_helper(lexer.token(), "SNIPPET_LINE_RANGE", "SnippetLineRange", 10)
    token_assert_helper(lexer.token(), "LINE", "5:23", 10)


def test_tokenization_of_annotation(lexer):
    annotation_str = "\n".join(
        [
            "Annotator: Person: Jane Doe()",
            "AnnotationDate: 2010-01-29T18:30:22Z",
            "AnnotationComment: <text>Document level annotation</text>",
            "AnnotationType: OTHER",
            f"SPDXREF: {DOCUMENT_SPDX_ID}",
        ]
    )

    lexer.input(annotation_str)
    token_assert_helper(lexer.token(), "ANNOTATOR", "Annotator", 1)
    token_assert_helper(lexer.token(), "PERSON_VALUE", "Person: Jane Doe()", 1)
    token_assert_helper(lexer.token(), "ANNOTATION_DATE", "AnnotationDate", 2)
    token_assert_helper(lexer.token(), "ISO8601_DATE", "2010-01-29T18:30:22Z", 2)
    token_assert_helper(lexer.token(), "ANNOTATION_COMMENT", "AnnotationComment", 3)
    token_assert_helper(lexer.token(), "TEXT", "<text>Document level annotation</text>", 3)
    token_assert_helper(lexer.token(), "ANNOTATION_TYPE", "AnnotationType", 4)
    token_assert_helper(lexer.token(), "LINE", "OTHER", 4)
    token_assert_helper(lexer.token(), "ANNOTATION_SPDX_ID", "SPDXREF", 5)
    token_assert_helper(lexer.token(), "LINE", DOCUMENT_SPDX_ID, 5)


def test_tokenization_of_relationship(lexer):
    relationship_str = "\n".join(
        [
            f"Relationship: {DOCUMENT_SPDX_ID} DESCRIBES NONE",
            "RelationshipComment: This is a comment.",
            "Relationship: DocumentRef-extern:SPDXRef-Package DESCRIBES NONE",
        ]
    )

    lexer.input(relationship_str)
    token_assert_helper(lexer.token(), "RELATIONSHIP", "Relationship", 1)
    token_assert_helper(lexer.token(), "LINE", f"{DOCUMENT_SPDX_ID} DESCRIBES NONE", 1)
    token_assert_helper(lexer.token(), "RELATIONSHIP_COMMENT", "RelationshipComment", 2)
    token_assert_helper(lexer.token(), "LINE", "This is a comment.", 2)
    token_assert_helper(lexer.token(), "RELATIONSHIP", "Relationship", 3)
    token_assert_helper(lexer.token(), "LINE", "DocumentRef-extern:SPDXRef-Package DESCRIBES NONE", 3)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Dict, Optional, Tuple
from rdflib import RDF, RDFS, BNode, Graph, Literal, URIRef

from spdx_tools.spdx.model import Snippet
from spdx_tools.spdx.rdfschema.namespace import POINTER_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.license_expression_writer import add_license_expression_or_none_or_no_assertion
from spdx_tools.spdx.writer.rdf.writer_utils import add_namespace_to_spdx_id, add_optional_literal


def add_snippet_to_graph(
    snippet: Snippet, graph: Graph, doc_namespace: str, external_doc_ref_to_namespace: Dict[str, str]
):
    snippet_resource = URIRef(add_namespace_to_spdx_id(snippet.spdx_id, doc_namespace, external_doc_ref_to_namespace))
    graph.add((snippet_resource, RDF.type, SPDX_NAMESPACE.Snippet))

    snippet_from_file_ref = URIRef(
        add_namespace_to_spdx_id(snippet.file_spdx_id, doc_namespace, external_doc_ref_to_namespace)
    )
    graph.add((snippet_resource, SPDX_NAMESPACE.snippetFromFile, snippet_from_file_ref))
    add_range_to_graph(
        snippet.byte_range, graph, snippet_resource, snippet_from_file_ref, POINTER_NAMESPACE.ByteOffsetPointer
    )
    add_range_to_graph(
        snippet.line_range, graph, snippet_resource, snippet_from_file_ref, POINTER_NAMESPACE.LineCharPointer
    )
    add_license_expression_or_none_or_no_assertion(
        snippet.license_concluded, graph, snippet_resource, SPDX_NAMESPACE.licenseConcluded, doc_namespace
    )
    add_license_expression_or_none_or_no_assertion(
        snippet.license_info_in_snippet, graph, snippet_resource, SPDX_NAMESPACE.licenseInfoInSnippet, doc_namespace
    )
    add_optional_literal(snippet.license_comment, graph, snippet_resource, SPDX_NAMESPACE.licenseComments)
    add_optional_literal(snippet.copyright_text, graph, snippet_resource, SPDX_NAMESPACE.copyrightText)
    add_optional_literal(snippet.comment, graph, snippet_resource, RDFS.comment)
    add_optional_literal(snippet.name, graph, snippet_resource, SPDX_NAMESPACE.name)
    for attribution_text in snippet.attribution_texts:
        graph.add((snippet_resource, SPDX_NAMESPACE.attributionText, Literal(attribution_text)))


def add_range_to_graph(
    range_information: Optional[Tuple[int, int]],
    graph: Graph,
    snippet_node: URIRef,
    snippet_from_file_ref: URIRef,
    pointer_class: URIRef,
):
    start_end_pointer = BNode()
    graph.add((start_end_pointer, RDF.type, POINTER_NAMESPACE.StartEndPointer))
    for predicate, value in [
        (POINTER_NAMESPACE.startPointer, range_information[0]),
        (POINTER_NAMESPACE.endPointer, range_information[1]),
    ]:
        pointer_node = BNode()
        graph.add((pointer_node, RDF.type, pointer_class))
        graph.add((start_end_pointer, predicate, pointer_node))
        graph.add((pointer_node, POINTER_NAMESPACE.reference, snippet_from_file_ref))
        if pointer_class == POINTER_NAMESPACE.ByteOffsetPointer:
            graph.add((pointer_node, POINTER_NAMESPACE.offset, Literal(value)))
        else:
            graph.add((pointer_node, POINTER_NAMESPACE.lineNumber, Literal(value)))

    graph.add((snippet_node, SPDX_NAMESPACE.range, start_end_pointer))



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.jsonschema.relationship_properties import RelationshipProperty
from spdx_tools.spdx.model import Document, Relationship


class RelationshipConverter(TypedConverter[Relationship]):
    def _get_property_value(
        self, relationship: Relationship, relationship_property: RelationshipProperty, document: Document = None
    ) -> Any:
        if relationship_property == RelationshipProperty.SPDX_ELEMENT_ID:
            return relationship.spdx_element_id
        elif relationship_property == RelationshipProperty.COMMENT:
            return relationship.comment
        elif relationship_property == RelationshipProperty.RELATED_SPDX_ELEMENT:
            return str(relationship.related_spdx_element_id)
        elif relationship_property == RelationshipProperty.RELATIONSHIP_TYPE:
            return relationship.relationship_type.name

    def get_json_type(self) -> Type[JsonProperty]:
        return RelationshipProperty

    def get_data_model_type(self) -> Type[Relationship]:
        return Relationship



#  SPDX-FileCopyrightText: 2023 spdx contributors
#
#  SPDX-License-Identifier: Apache-2.0
import os.path
import runpy

import pytest


@pytest.fixture
def cleanup_output_files():
    yield

    files_to_delete = ["spdx2_to_3.jsonld", "my_spdx_document.spdx.json", "converted_format.xml", "graph.png"]
    for file in files_to_delete:
        if os.path.exists(file):
            os.remove(file)


def run_example(example_file: str):
    file_path = os.path.join(os.path.dirname(__file__), "../../../examples/", example_file)
    runpy.run_path(file_path)


def test_spdx2_parse_file():
    run_example("spdx2_parse_file.py")


@pytest.mark.usefixtures("cleanup_output_files")
def test_spdx2_convert_to_spdx3():
    run_example("spdx2_convert_to_spdx3.py")
    assert os.path.exists("spdx2_to_3.jsonld")


@pytest.mark.usefixtures("cleanup_output_files")
def test_spdx2_document_from_scratch():
    run_example("spdx2_document_from_scratch.py")
    assert os.path.exists("my_spdx_document.spdx.json")


@pytest.mark.usefixtures("cleanup_output_files")
def test_spdx2_convert_format():
    run_example("spdx2_convert_format.py")
    assert os.path.exists("converted_format.xml")


@pytest.mark.usefixtures("cleanup_output_files")
def test_spdx2_generate_graph():
    try:
        import networkx  # noqa F401
        import pygraphviz  # noqa F401
    except ImportError:
        pytest.skip("Missing optional dependencies")

    run_example("spdx2_generate_graph.py")
    assert os.path.exists("graph.png")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.validation.creation_info_validator import validate_creation_info
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import creation_info_fixture


def test_valid_creation_info():
    creation_info = creation_info_fixture()
    validation_messages: List[ValidationMessage] = validate_creation_info(creation_info, "SPDX-2.3")

    assert validation_messages == []


@pytest.mark.parametrize(
    "creation_info_input, spdx_id, expected_message",
    [
        (
            creation_info_fixture(spdx_id="SPDXRef-doc"),
            "SPDXRef-doc",
            f"spdx_id must be {DOCUMENT_SPDX_ID}, but is: SPDXRef-doc",
        ),
        (creation_info_fixture(data_license="MIT"), DOCUMENT_SPDX_ID, 'data_license must be "CC0-1.0", but is: MIT'),
        (
            creation_info_fixture(document_namespace="some_namespace"),
            DOCUMENT_SPDX_ID,
            "document_namespace must be a valid URI specified in RFC-3986 and must contain no fragment (#), "
            "but is: some_namespace",
        ),
    ],
)
def test_invalid_creation_info(creation_info_input, expected_message, spdx_id):
    validation_messages: List[ValidationMessage] = validate_creation_info(creation_info_input, "SPDX-2.3")

    expected = ValidationMessage(expected_message, ValidationContext(spdx_id, None, SpdxElementType.DOCUMENT))

    assert validation_messages == [expected]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDF, RDFS, BNode, Graph, Literal, URIRef

from spdx_tools.spdx.datetime_conversions import datetime_to_iso_string
from spdx_tools.spdx.model import CreationInfo
from spdx_tools.spdx.rdfschema.namespace import LICENSE_NAMESPACE, SPDX_NAMESPACE
from spdx_tools.spdx.writer.rdf.external_document_ref_writer import add_external_document_ref_to_graph
from spdx_tools.spdx.writer.rdf.writer_utils import add_optional_literal


def add_creation_info_to_graph(creation_info: CreationInfo, graph: Graph):
    doc_node = URIRef(f"{creation_info.document_namespace}#{creation_info.spdx_id}")
    graph.add((doc_node, RDF.type, SPDX_NAMESPACE.SpdxDocument))
    graph.add((doc_node, SPDX_NAMESPACE.specVersion, Literal(creation_info.spdx_version)))
    graph.add((doc_node, SPDX_NAMESPACE.dataLicense, LICENSE_NAMESPACE[creation_info.data_license]))
    graph.add((doc_node, SPDX_NAMESPACE.name, Literal(creation_info.name)))
    add_optional_literal(creation_info.document_comment, graph, doc_node, RDFS.comment)

    creation_info_node = BNode()
    graph.add((creation_info_node, RDF.type, SPDX_NAMESPACE.CreationInfo))

    graph.add((creation_info_node, SPDX_NAMESPACE.created, Literal(datetime_to_iso_string(creation_info.created))))

    for creator in creation_info.creators:
        graph.add((creation_info_node, SPDX_NAMESPACE.creator, Literal(creator.to_serialized_string())))

    add_optional_literal(
        creation_info.license_list_version, graph, creation_info_node, SPDX_NAMESPACE.licenseListVersion
    )
    add_optional_literal(creation_info.creator_comment, graph, creation_info_node, RDFS.comment)

    graph.add((doc_node, SPDX_NAMESPACE.creationInfo, creation_info_node))

    for external_document_ref in creation_info.external_document_refs:
        add_external_document_ref_to_graph(external_document_ref, graph, doc_node, creation_info.document_namespace)

    return doc_node



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import List


class Logger:
    messages: List[str]

    def __init__(self):
        self.messages = []

    def append(self, message: str):
        self.messages.append(message)

    def extend(self, messages_to_append: List[str]):
        self.messages.extend(messages_to_append)

    def has_messages(self):
        return bool(self.messages)

    def get_messages(self):
        return list(self.messages)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Type

from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.external_package_ref_properties import ExternalPackageRefProperty
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.model import Document, ExternalPackageRef


class ExternalPackageRefConverter(TypedConverter[ExternalPackageRef]):
    def _get_property_value(
        self,
        external_ref: ExternalPackageRef,
        external_ref_property: ExternalPackageRefProperty,
        document: Document = None,
    ) -> Any:
        if external_ref_property == ExternalPackageRefProperty.COMMENT:
            return external_ref.comment
        elif external_ref_property == ExternalPackageRefProperty.REFERENCE_CATEGORY:
            return external_ref.category.name
        elif external_ref_property == ExternalPackageRefProperty.REFERENCE_LOCATOR:
            return external_ref.locator
        elif external_ref_property == ExternalPackageRefProperty.REFERENCE_TYPE:
            return external_ref.reference_type

    def get_json_type(self) -> Type[JsonProperty]:
        return ExternalPackageRefProperty

    def get_data_model_type(self) -> Type[ExternalPackageRef]:
        return ExternalPackageRef



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model import ExternalIdentifier, ExternalIdentifierType
from tests.spdx3.fixtures import external_identifier_fixture
from tests.spdx3.model.model_test_utils import get_property_names


def test_correct_initialization():
    external_identifier = external_identifier_fixture()

    for property_name in get_property_names(ExternalIdentifier):
        assert getattr(external_identifier, property_name) is not None

    assert external_identifier.external_identifier_type == ExternalIdentifierType.OTHER
    assert external_identifier.identifier == "externalIdentifierIdentifier"
    assert external_identifier.comment == "externalIdentifierComment"
    assert external_identifier.identifier_locator == [
        "https://spdx.test/tools-python/external_identifier_identifier_locator"
    ]
    assert (
        external_identifier.issuing_authority == "https://spdx.test/tools-python/external_identifier_issuing_authority"
    )


def test_invalid_initialization():
    with pytest.raises(TypeError) as err:
        ExternalIdentifier("CPE22", ["identifier", "another_identifier"], 34, "locator", True)

    assert len(err.value.args[0]) == 5
    for error in err.value.args[0]:
        assert error.startswith("SetterError ExternalIdentifier:")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class CreationInfoProperty(JsonProperty):
    CREATED = auto()
    CREATORS = auto()
    LICENSE_LIST_VERSION = auto()
    COMMENT = auto()



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import abstractmethod
from dataclasses import field

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.spdx3.model.licensing.any_license_info import AnyLicenseInfo


@dataclass_with_properties
class License(AnyLicenseInfo):
    license_id: str
    license_name: str
    license_text: str
    license_comment: Optional[str] = None
    see_also: List[str] = field(default_factory=list)
    is_osi_approved: Optional[bool] = None
    is_fsf_libre: Optional[bool] = None
    standard_license_header: Optional[str] = None
    standard_license_template: Optional[str] = None
    is_deprecated_license_id: Optional[bool] = None
    obsoleted_by: Optional[str] = None

    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from spdx_tools.spdx3.bump_from_spdx2.annotation import bump_annotation
from spdx_tools.spdx3.bump_from_spdx2.creation_info import bump_creation_info
from spdx_tools.spdx3.bump_from_spdx2.file import bump_file
from spdx_tools.spdx3.bump_from_spdx2.package import bump_package
from spdx_tools.spdx3.bump_from_spdx2.relationship import bump_relationships
from spdx_tools.spdx3.bump_from_spdx2.snippet import bump_snippet
from spdx_tools.spdx3.model import CreationInfo, SpdxDocument
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx.model import RelationshipType
from spdx_tools.spdx.model.document import Document as Spdx2_Document
from spdx_tools.spdx.model.relationship_filters import filter_by_type_and_origin

""" We want to implement a bump_from_spdx2 from the data model in src.spdx to the data model in src.spdx3.
    As there are many fundamental differences between these version we want each bump_from_spdx2 method to take
    the object from src.spdx and add all objects that the input is translated to into the payload."""


def bump_spdx_document(document: Spdx2_Document) -> Payload:
    payload = Payload()
    document_namespace: str = document.creation_info.document_namespace
    spdx_document: SpdxDocument = bump_creation_info(document.creation_info, payload)
    spdx_document.root_element = [
        f"{document_namespace}#{relationship.related_spdx_element_id}"
        for relationship in filter_by_type_and_origin(
            document.relationships, RelationshipType.DESCRIBES, "SPDXRef-DOCUMENT"
        )
    ]

    creation_info: CreationInfo = spdx_document.creation_info

    payload.add_element(spdx_document)

    for spdx2_package in document.packages:
        bump_package(
            spdx2_package,
            payload,
            document_namespace,
            document.creation_info.external_document_refs,
            spdx_document.imports,
        )

    for spdx2_file in document.files:
        bump_file(
            spdx2_file,
            payload,
            document_namespace,
            document.creation_info.external_document_refs,
            spdx_document.imports,
        )

    for spdx2_snippet in document.snippets:
        bump_snippet(
            spdx2_snippet,
            payload,
            document_namespace,
            document.creation_info.external_document_refs,
            spdx_document.imports,
        )

    bump_relationships(document.relationships, payload, document_namespace)

    for counter, spdx2_annotation in enumerate(document.annotations):
        bump_annotation(spdx2_annotation, payload, creation_info, document_namespace, counter)

    spdx_document.element = [spdx_id for spdx_id in payload.get_full_map() if spdx_id != spdx_document.spdx_id]

    return payload



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDFS, BNode, Graph, URIRef

from spdx_tools.spdx.datetime_conversions import datetime_from_str
from spdx_tools.spdx.model import Annotation, AnnotationType
from spdx_tools.spdx.parser.actor_parser import ActorParser
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import parse_enum_value, parse_literal, parse_spdx_id
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def parse_annotation(annotation_node: BNode, graph: Graph, parent_node: URIRef, doc_namespace: str) -> Annotation:
    logger = Logger()
    spdx_id = parse_spdx_id(parent_node, doc_namespace, graph)
    annotator = parse_literal(
        logger, graph, annotation_node, SPDX_NAMESPACE.annotator, parsing_method=ActorParser.parse_actor
    )
    annotation_type = parse_literal(
        logger,
        graph,
        annotation_node,
        SPDX_NAMESPACE.annotationType,
        parsing_method=lambda x: parse_enum_value(x, AnnotationType, SPDX_NAMESPACE.annotationType_),
    )
    annotation_date = parse_literal(
        logger, graph, annotation_node, SPDX_NAMESPACE.annotationDate, parsing_method=datetime_from_str
    )
    annotation_comment = parse_literal(logger, graph, annotation_node, RDFS.comment)

    raise_parsing_error_if_logger_has_messages(logger, "Annotation")
    annotation = construct_or_raise_parsing_error(
        Annotation,
        dict(
            spdx_id=spdx_id,
            annotation_type=annotation_type,
            annotator=annotator,
            annotation_date=annotation_date,
            annotation_comment=annotation_comment,
        ),
    )

    return annotation



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDFS, Graph, URIRef

from spdx_tools.spdx.model import ExtractedLicensingInfo
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import (
    get_correctly_typed_triples,
    parse_literal,
    parse_literal_or_no_assertion_or_none,
)
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def parse_extracted_licensing_info(
    extracted_licensing_info_node: URIRef, graph: Graph, doc_namespace: str
) -> ExtractedLicensingInfo:
    logger = Logger()
    license_id = parse_literal(logger, graph, extracted_licensing_info_node, SPDX_NAMESPACE.licenseId)
    if not license_id:
        license_id = (
            extracted_licensing_info_node.fragment
            if extracted_licensing_info_node.startswith(f"{doc_namespace}#")
            else extracted_licensing_info_node.toPython()
        )

    extracted_text = parse_literal(logger, graph, extracted_licensing_info_node, SPDX_NAMESPACE.extractedText)
    comment = parse_literal(logger, graph, extracted_licensing_info_node, RDFS.comment)
    license_name = parse_literal_or_no_assertion_or_none(
        logger, graph, extracted_licensing_info_node, SPDX_NAMESPACE.name
    )
    cross_references = []
    for _, _, cross_reference_node in get_correctly_typed_triples(
        logger, graph, extracted_licensing_info_node, RDFS.seeAlso
    ):
        cross_references.append(cross_reference_node.toPython())
    raise_parsing_error_if_logger_has_messages(logger, "ExtractedLicensingInfo")
    extracted_licensing_info = construct_or_raise_parsing_error(
        ExtractedLicensingInfo,
        dict(
            license_id=license_id,
            extracted_text=extracted_text,
            comment=comment,
            license_name=license_name,
            cross_references=cross_references,
        ),
    )

    return extracted_licensing_info



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List

from spdx_tools.spdx.model import Annotation, Document
from spdx_tools.spdx.validation.actor_validator import validate_actor
from spdx_tools.spdx.validation.spdx_id_validators import validate_spdx_id
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_annotations(annotations: List[Annotation], document: Document) -> List[ValidationMessage]:
    validation_messages = []
    for annotation in annotations:
        validation_messages.extend(validate_annotation(annotation, document))

    return validation_messages


def validate_annotation(annotation: Annotation, document: Document) -> List[ValidationMessage]:
    validation_messages = []
    context = ValidationContext(element_type=SpdxElementType.ANNOTATION, full_element=annotation)

    validation_messages.extend(validate_actor(annotation.annotator, "annotation"))

    messages: List[str] = validate_spdx_id(annotation.spdx_id, document, check_document=True)
    for message in messages:
        validation_messages.append(ValidationMessage(message, context))

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from dataclasses import field

from beartype.typing import List, Optional

from spdx_tools.common.typing.dataclass_with_properties import dataclass_with_properties
from spdx_tools.spdx3.model import CreationInfo, ExternalIdentifier, ExternalReference, IntegrityMethod


@dataclass_with_properties
class Element(ABC):
    spdx_id: str  # IRI
    creation_info: Optional[CreationInfo] = None
    name: Optional[str] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    comment: Optional[str] = None
    verified_using: List[IntegrityMethod] = field(default_factory=list)
    external_reference: List[ExternalReference] = field(default_factory=list)
    external_identifier: List[ExternalIdentifier] = field(default_factory=list)
    extension: Optional[str] = None  # placeholder for extension

    @abstractmethod
    def __init__(self):
        pass



# SPDX-License-Identifier: Apache-2.0
#  Copyright (c) 2022 spdx contributors
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#    http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from beartype.typing import Any, Callable, Dict, List, Optional, TextIO, Tuple, Union
from license_expression import LicenseExpression

from spdx_tools.spdx.model import (
    Actor,
    File,
    Package,
    Relationship,
    RelationshipType,
    Snippet,
    SpdxNoAssertion,
    SpdxNone,
)


def write_separator(out: TextIO):
    out.write("\n")


def write_value(
    tag: str, value: Optional[Union[bool, str, SpdxNone, SpdxNoAssertion, LicenseExpression]], out: TextIO
):
    if value is not None:
        out.write(f"{tag}: {value}\n")


def write_range(tag: str, value: Optional[Tuple[int, int]], out: TextIO):
    if value:
        out.write(f"{tag}: {value[0]}:{value[1]}\n")


def write_text_value(tag: str, value: Optional[Union[str, SpdxNone, SpdxNoAssertion]], out: TextIO):
    if isinstance(value, str) and "\n" in value:
        out.write(f"{tag}: <text>{value}</text>\n")
    else:
        write_value(tag, value, out)


def transform_enum_name_to_tv(enum_str: str) -> str:
    return enum_str.replace("_", "-")


def write_optional_heading(optional_field: Any, heading: str, text_output: TextIO):
    if optional_field:
        text_output.write(heading)


def write_list_of_elements(
    list_of_elements: List[Any],
    write_method: Callable[[Any, TextIO], None],
    text_output: TextIO,
    with_separator: bool = False,
):
    for element in list_of_elements:
        write_method(element, text_output)
        if with_separator:
            write_separator(text_output)


def write_actor(tag: str, element_to_write: Optional[Union[Actor, SpdxNoAssertion]], text_output: TextIO):
    if isinstance(element_to_write, Actor):
        write_value(tag, element_to_write.to_serialized_string(), text_output)
    else:
        write_value(tag, element_to_write, text_output)


def scan_relationships(
    relationships: List[Relationship], packages: List[Package], files: List[File]
) -> Tuple[List, Dict]:
    contained_files_by_package_id = dict()
    relationships_to_write = []
    files_by_spdx_id = {file.spdx_id: file for file in files}
    packages_spdx_ids = [package.spdx_id for package in packages]
    for relationship in relationships:
        if relationship.related_spdx_element_id in [SpdxNoAssertion(), SpdxNone()]:
            relationships_to_write.append(relationship)
        elif (
            relationship.relationship_type == RelationshipType.CONTAINS
            and relationship.spdx_element_id in packages_spdx_ids
            and relationship.related_spdx_element_id in files_by_spdx_id.keys()
        ):
            contained_files_by_package_id.setdefault(relationship.spdx_element_id, []).append(
                files_by_spdx_id[relationship.related_spdx_element_id]
            )
            if relationship.comment:
                relationships_to_write.append(relationship)
        elif (
            relationship.relationship_type == RelationshipType.CONTAINED_BY
            and relationship.related_spdx_element_id in packages_spdx_ids
            and relationship.spdx_element_id in files_by_spdx_id
        ):
            contained_files_by_package_id.setdefault(relationship.related_spdx_element_id, []).append(
                files_by_spdx_id[relationship.spdx_element_id]
            )
            if relationship.comment:
                relationships_to_write.append(relationship)
        else:
            relationships_to_write.append(relationship)

    return relationships_to_write, contained_files_by_package_id


def get_file_ids_with_contained_snippets(snippets: List[Snippet], files: List[File]) -> Dict:
    file_ids_with_contained_snippets = dict()
    file_spdx_ids: List[str] = [file.spdx_id for file in files]
    for snippet in snippets:
        if snippet.file_spdx_id in file_spdx_ids:
            file_ids_with_contained_snippets.setdefault(snippet.file_spdx_id, []).append(snippet)

    return file_ids_with_contained_snippets



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model import Artifact, Element, ElementCollection, IntegrityMethod


@pytest.mark.parametrize("abstract_class", [Element, Artifact, ElementCollection, IntegrityMethod])
def test_initialization_throws_error(abstract_class):
    with pytest.raises(TypeError) as err:
        abstract_class()

    assert f"Can't instantiate abstract class {abstract_class.__name__}" in err.value.args[0]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os

import pytest
from rdflib import BNode, Graph, URIRef

from spdx_tools.spdx.model import ChecksumAlgorithm
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.rdf.checksum_parser import convert_rdf_to_algorithm, parse_checksum
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def test_parse_checksum():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    checksum_node = graph.value(
        subject=URIRef("https://some.namespace#DocumentRef-external"), predicate=SPDX_NAMESPACE.checksum
    )
    assert isinstance(checksum_node, BNode)

    checksum = parse_checksum(checksum_node, graph)

    assert checksum.algorithm == ChecksumAlgorithm.SHA1
    assert checksum.value == "71c4025dd9897b364f3ebbb42c484ff43d00791c"


@pytest.mark.parametrize(
    "rdf_element,expected",
    [
        (SPDX_NAMESPACE.checksumAlgorithm_sha1, ChecksumAlgorithm.SHA1),
        (SPDX_NAMESPACE.checksumAlgorithm_sha224, ChecksumAlgorithm.SHA224),
        (SPDX_NAMESPACE.checksumAlgorithm_sha256, ChecksumAlgorithm.SHA256),
        (SPDX_NAMESPACE.checksumAlgorithm_sha384, ChecksumAlgorithm.SHA384),
        (SPDX_NAMESPACE.checksumAlgorithm_sha512, ChecksumAlgorithm.SHA512),
        (SPDX_NAMESPACE.checksumAlgorithm_sha3_256, ChecksumAlgorithm.SHA3_256),
        (SPDX_NAMESPACE.checksumAlgorithm_sha3_384, ChecksumAlgorithm.SHA3_384),
        (SPDX_NAMESPACE.checksumAlgorithm_sha3_512, ChecksumAlgorithm.SHA3_512),
        (SPDX_NAMESPACE.checksumAlgorithm_blake2b256, ChecksumAlgorithm.BLAKE2B_256),
        (SPDX_NAMESPACE.checksumAlgorithm_blake2b384, ChecksumAlgorithm.BLAKE2B_384),
        (SPDX_NAMESPACE.checksumAlgorithm_blake2b512, ChecksumAlgorithm.BLAKE2B_512),
        (SPDX_NAMESPACE.checksumAlgorithm_blake3, ChecksumAlgorithm.BLAKE3),
        (SPDX_NAMESPACE.checksumAlgorithm_md2, ChecksumAlgorithm.MD2),
        (SPDX_NAMESPACE.checksumAlgorithm_md4, ChecksumAlgorithm.MD4),
        (SPDX_NAMESPACE.checksumAlgorithm_md5, ChecksumAlgorithm.MD5),
        (SPDX_NAMESPACE.checksumAlgorithm_md6, ChecksumAlgorithm.MD6),
        (SPDX_NAMESPACE.checksumAlgorithm_adler32, ChecksumAlgorithm.ADLER32),
    ],
)
def test_convert_rdf_to_algorithm(rdf_element, expected):
    algorithm = convert_rdf_to_algorithm(rdf_element)

    assert algorithm == expected


@pytest.mark.parametrize(
    "invalid_rdf_element",
    [
        SPDX_NAMESPACE.checksumAlgorithm_blake2b_512,
        SPDX_NAMESPACE.checksumAlgorithm_BLAKE2b_512,
        SPDX_NAMESPACE.checksumAlgorithm_sha3512,
        SPDX_NAMESPACE.checksumAlgorithm_sha513,
        SPDX_NAMESPACE.checksumalgorithm_blake2b_512,
    ],
)
def test_convert_invalid_rdf_to_algorithm(invalid_rdf_element):
    with pytest.raises(SPDXParsingError):
        convert_rdf_to_algorithm(invalid_rdf_element)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os
from unittest import TestCase

import pytest
from rdflib import RDF, BNode, Graph, Literal, URIRef

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Checksum,
    ChecksumAlgorithm,
    ExternalPackageRefCategory,
    PackagePurpose,
    PackageVerificationCode,
    SpdxNoAssertion,
)
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.rdf.package_parser import parse_external_package_ref, parse_package
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def test_package_parser():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    # we have two packages in the test file, graph.value() will return the first package
    package_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.Package)
    doc_namespace = "https://some.namespace"
    assert isinstance(package_node, URIRef)

    package = parse_package(package_node, graph, doc_namespace)

    assert package.spdx_id == "SPDXRef-Package"
    assert package.name == "packageName"
    assert package.download_location == "https://download.com"
    assert package.version == "12.2"
    assert package.file_name == "./packageFileName"
    assert package.homepage == "https://homepage.com"
    assert package.files_analyzed is True
    assert package.checksums == [Checksum(ChecksumAlgorithm.SHA1, "71c4025dd9897b364f3ebbb42c484ff43d00791c")]
    assert package.source_info == "sourceInfo"
    assert package.license_concluded == spdx_licensing.parse("MIT AND GPL-2.0")
    assert package.license_declared == spdx_licensing.parse("MIT AND GPL-2.0")
    TestCase().assertCountEqual(
        package.license_info_from_files,
        [spdx_licensing.parse("MIT"), spdx_licensing.parse("GPL-2.0"), SpdxNoAssertion()],
    )
    assert package.license_comment == "packageLicenseComment"
    assert package.copyright_text == "packageCopyrightText"
    assert package.verification_code == PackageVerificationCode(
        value="85ed0817af83a24ad8da68c2b5094de69833983c", excluded_files=["./exclude.py"]
    )
    assert len(package.external_references) == 1
    assert package.summary == "packageSummary"
    assert package.description == "packageDescription"
    assert package.comment == "packageComment"
    assert package.attribution_texts == ["packageAttributionText"]
    assert package.primary_package_purpose == PackagePurpose.SOURCE
    assert package.supplier == Actor(ActorType.PERSON, "supplierName", "some@mail.com")
    assert package.originator == Actor(ActorType.PERSON, "originatorName", "some@mail.com")


@pytest.mark.parametrize(
    "download_location,category,locator,type,comment",
    [
        (
            "https://download.com",
            ExternalPackageRefCategory.PACKAGE_MANAGER,
            "org.apache.tomcat:tomcat:9.0.0.M4",
            "maven-central",
            "externalPackageRefComment",
        ),
        (
            "http://differentdownload.com",
            ExternalPackageRefCategory.OTHER,
            "acmecorp/acmenator/4.1.3-alpha",
            "LocationRef-acmeforge",
            "This is the external ref for Acme",
        ),
    ],
)
def test_external_package_ref_parser(download_location, category, locator, type, comment):
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    doc_namespace = "https://some.namespace"
    # we use the download location to identify the package node
    # in the test file we have two different external package refs depending on the package
    package_node = graph.value(predicate=SPDX_NAMESPACE.downloadLocation, object=Literal(download_location))
    external_package_ref_node = graph.value(package_node, SPDX_NAMESPACE.externalRef)
    assert isinstance(external_package_ref_node, BNode)

    external_package_ref = parse_external_package_ref(external_package_ref_node, graph, doc_namespace)

    assert external_package_ref.category == category
    assert external_package_ref.locator == locator
    assert external_package_ref.reference_type == type
    assert external_package_ref.comment == comment


def test_parse_invalid_package():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/invalid_documents/file_without_spdx_ids.xml"))
    package_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.Package)
    doc_namespace = "https://some.namespace"

    assert isinstance(package_node, BNode)
    with pytest.raises(SPDXParsingError):
        parse_package(package_node, graph, doc_namespace)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx.casing_tools import camel_case_to_snake_case, snake_case_to_camel_case


@pytest.mark.parametrize("snake_case_str,camel_case_str", [("snake_case", "snakeCase")])
def test_snake_case_to_camel_case(snake_case_str, camel_case_str):
    camel_case = snake_case_to_camel_case(snake_case_str)

    assert camel_case == camel_case_str


@pytest.mark.parametrize(
    "camel_case_str,snake_case_str",
    [("camelCase", "camel_case"), ("camelCaseMore", "camel_case_more"), ("CamelCase", "camel_case")],
)
def test_camel_case_to_snake_case(camel_case_str, snake_case_str):
    snake_case = camel_case_to_snake_case(camel_case_str)

    assert snake_case == snake_case_str



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from beartype.typing import List

from spdx_tools.spdx.model import Document, Relationship, RelationshipType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.validation.spdx_id_validators import validate_spdx_id
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage


def validate_relationships(
    relationships: List[Relationship], spdx_version: str, document: Document
) -> List[ValidationMessage]:
    validation_messages = []
    for relationship in relationships:
        validation_messages.extend(validate_relationship(relationship, spdx_version, document))

    return validation_messages


def validate_relationship(
    relationship: Relationship, spdx_version: str, document: Document
) -> List[ValidationMessage]:
    validation_messages = []
    context = ValidationContext(element_type=SpdxElementType.RELATIONSHIP, full_element=relationship)

    relationship_type: RelationshipType = relationship.relationship_type

    messages: List[str] = validate_spdx_id(relationship.spdx_element_id, document, check_document=True)
    for message in messages:
        validation_messages.append(ValidationMessage(message, context))

    if relationship.related_spdx_element_id not in [SpdxNone(), SpdxNoAssertion()]:
        messages: List[str] = validate_spdx_id(relationship.related_spdx_element_id, document, check_document=True)
        for message in messages:
            validation_messages.append(ValidationMessage(message, context))

    if spdx_version == "SPDX-2.2":
        if (
            relationship_type == RelationshipType.SPECIFICATION_FOR
            or relationship_type == RelationshipType.REQUIREMENT_DESCRIPTION_FOR
        ):
            validation_messages.append(ValidationMessage(f"{relationship_type} is not supported in SPDX-2.2", context))

    return validation_messages



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import AnnotationType
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.tagvalue.parser import Parser
from tests.spdx.parser.tagvalue.test_creation_info_parser import DOCUMENT_STR


def test_parse_annotation():
    parser = Parser()
    annotation_str = "\n".join(
        [
            "Annotator: Person: Jane Doe()",
            "AnnotationDate: 2010-01-29T18:30:22Z",
            "AnnotationComment: <text>Document level annotation</text>",
            "AnnotationType: OTHER",
            f"SPDXREF: {DOCUMENT_SPDX_ID}",
        ]
    )
    document = parser.parse("\n".join([DOCUMENT_STR, annotation_str]))
    assert document is not None
    assert len(document.annotations) == 1
    annotation = document.annotations[0]
    assert annotation.annotator.name == "Jane Doe"
    assert annotation.annotation_date == datetime(2010, 1, 29, 18, 30, 22)
    assert annotation.annotation_comment == "Document level annotation"
    assert annotation.annotation_type == AnnotationType.OTHER
    assert annotation.spdx_id == DOCUMENT_SPDX_ID


def test_parse_annotation_with_organization_as_annotator():
    parser = Parser()
    annotation_str = "\n".join(
        [
            "Annotator: Organization: some-organization",
            "AnnotationDate: 2010-01-29T18:30:22Z",
            "AnnotationComment: <text>Document level annotation</text>",
            "AnnotationType: OTHER",
            f"SPDXREF: {DOCUMENT_SPDX_ID}",
        ]
    )
    document = parser.parse("\n".join([DOCUMENT_STR, annotation_str]))
    assert document is not None
    assert len(document.annotations) == 1
    annotation = document.annotations[0]
    assert annotation.annotator.name == "some-organization"
    assert annotation.annotation_date == datetime(2010, 1, 29, 18, 30, 22)
    assert annotation.annotation_comment == "Document level annotation"
    assert annotation.annotation_type == AnnotationType.OTHER
    assert annotation.spdx_id == DOCUMENT_SPDX_ID


@pytest.mark.parametrize(
    "annotation_str, expected_message",
    [
        (
            "Annotator: Person: Jane Doe()",
            r"__init__() missing 4 "
            "required positional arguments: 'spdx_id', 'annotation_type', "
            "'annotation_date', and 'annotation_comment'",
        ),
        (
            "Annotator: Person: Jane Doe()\nAnnotationType: SOURCE\nAnnotationDate: 201001-2912:23",
            "Error while parsing Annotation: ['Invalid AnnotationType: SOURCE. Line: 2', "
            "'Error while parsing AnnotationDate: Token did not match specified grammar "
            "rule. Line: 3']",
        ),
        (
            "Annotator: Jane Doe()\nAnnotationDate: 201001-29T18:30:22Z\n"
            "AnnotationComment: <text>Document level annotation</text>\nAnnotationType: OTHER\n"
            f"SPDXREF: {DOCUMENT_SPDX_ID}",
            "Error while parsing Annotation: ['Error while parsing Annotator: Token did "
            "not match specified grammar rule. Line: 1', 'Error while parsing "
            "AnnotationDate: Token did not match specified grammar rule. Line: 2']",
        ),
        ("Annotator: Person: ()", "Error while parsing Annotation: [['No name for Actor provided: Person: ().']]"),
        (
            "AnnotationType: REVIEW",
            "Element Annotation is not the current element in scope, probably the "
            "expected tag to start the element (Annotator) is missing. Line: 1",
        ),
    ],
)
def test_parse_invalid_annotation(annotation_str, expected_message):
    parser = Parser()
    with pytest.raises(SPDXParsingError) as err:
        parser.parse(annotation_str)

    assert expected_message in err.value.get_messages()[0]



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import os
from unittest import TestCase

from rdflib import RDF, Graph

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.parser.rdf import rdf_parser
from spdx_tools.spdx.parser.rdf.license_expression_parser import parse_license_expression
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def test_license_expression_parser():
    graph = Graph().parse(os.path.join(os.path.dirname(__file__), "data/file_to_test_rdf_parser.rdf.xml"))
    package_node = graph.value(predicate=RDF.type, object=SPDX_NAMESPACE.Package)
    license_expression_node = graph.value(subject=package_node, predicate=SPDX_NAMESPACE.licenseConcluded)

    license_expression = parse_license_expression(license_expression_node, graph, "https://some.namespace#")

    assert license_expression == spdx_licensing.parse("GPL-2.0 AND MIT")


def test_license_expression_parser_with_coupled_licenses():
    doc = rdf_parser.parse_from_file(
        os.path.join(os.path.dirname(__file__), "../../data/SPDXRdfExample-v2.3.spdx.rdf.xml")
    )

    packages_by_spdx_id = {package.spdx_id: package for package in doc.packages}
    files_by_spdx_id = {file.spdx_id: file for file in doc.files}

    assert packages_by_spdx_id["SPDXRef-Package"].license_declared == spdx_licensing.parse(
        "LGPL-2.0-only AND LicenseRef-3"
    )
    assert packages_by_spdx_id["SPDXRef-Package"].license_concluded == spdx_licensing.parse(
        "LGPL-2.0-only OR LicenseRef-3"
    )
    TestCase().assertCountEqual(
        packages_by_spdx_id["SPDXRef-Package"].license_info_from_files,
        [
            spdx_licensing.parse("GPL-2.0"),
            spdx_licensing.parse("LicenseRef-1"),
            spdx_licensing.parse("LicenseRef-2"),
        ],
    )

    assert files_by_spdx_id["SPDXRef-JenaLib"].license_concluded == spdx_licensing.parse("LicenseRef-1")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod


class LicenseField(ABC):
    @abstractmethod
    def __init__(self):
        pass



# SPDX-FileCopyrightText: 2024 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import io
from datetime import datetime

from semantic_version import Version

from spdx_tools.spdx3.model import CreationInfo, ProfileIdentifierType, SpdxDocument
from spdx_tools.spdx3.writer.console.spdx_document_writer import write_spdx_document


def test_render_creation_info():
    fake_datetime = datetime(year=2024, month=1, day=1)
    spec_version = Version("3.0.0")
    creation_info = CreationInfo(
        spec_version=spec_version,
        created=fake_datetime,
        created_by=[],
        profile=[ProfileIdentifierType.SOFTWARE],
    )
    spdx_document = SpdxDocument(
        spdx_id="SPDXRef-FOO",
        name="BAR",
        element=[],
        root_element=[],
        creation_info=creation_info,
    )
    output_str = io.StringIO()
    write_spdx_document(spdx_document, text_output=output_str)

    assert (
        output_str.getvalue()
        == """\
## SPDX Document
SPDXID: SPDXRef-FOO
name: BAR
# Creation Information
  specVersion: 3.0.0
  created: 2024-01-01T00:00:00Z
  profile: SOFTWARE
  data license: CC0-1.0
elements: 
"""  # noqa: W291 # elements: are printed with a space
    )



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Any, Dict
from rdflib import RDF, Graph

from spdx_tools.spdx.model import Document, RelationshipType
from spdx_tools.spdx.parser.error import SPDXParsingError
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.annotation_parser import parse_annotation
from spdx_tools.spdx.parser.rdf.creation_info_parser import parse_creation_info
from spdx_tools.spdx.parser.rdf.extracted_licensing_info_parser import parse_extracted_licensing_info
from spdx_tools.spdx.parser.rdf.file_parser import parse_file
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import get_correctly_typed_triples
from spdx_tools.spdx.parser.rdf.package_parser import parse_package
from spdx_tools.spdx.parser.rdf.relationship_parser import parse_implicit_relationship, parse_relationship
from spdx_tools.spdx.parser.rdf.snippet_parser import parse_snippet
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def parse_from_file(file_name: str, encoding: str = "utf-8") -> Document:
    graph = Graph()
    with open(file_name, encoding=encoding) as file:
        graph.parse(file, format="xml")

    document: Document = translate_graph_to_document(graph)
    return document


def translate_graph_to_document(graph: Graph) -> Document:
    parsed_fields: Dict[str, Any] = dict()
    logger = Logger()
    creation_info, doc_node = parse_creation_info(graph)

    parsed_fields["creation_info"] = creation_info

    for element, triple, parsing_method in [
        ("packages", (None, RDF.type, SPDX_NAMESPACE.Package), parse_package),
        ("files", (None, RDF.type, SPDX_NAMESPACE.File), parse_file),
        ("snippets", (None, RDF.type, SPDX_NAMESPACE.Snippet), parse_snippet),
    ]:
        elements = []
        for element_node, _, _ in get_correctly_typed_triples(logger, graph, *triple):
            try:
                elements.append(parsing_method(element_node, graph, creation_info.document_namespace))
            except SPDXParsingError as err:
                logger.extend(err.get_messages())
        parsed_fields[element] = elements

    for element, triple, parsing_method in [
        ("annotations", (None, SPDX_NAMESPACE.annotation, None), parse_annotation),
        ("relationships", (None, SPDX_NAMESPACE.relationship, None), parse_relationship),
    ]:
        elements = []
        for parent_node, _, element_node in graph.triples(triple):
            try:
                elements.append(parsing_method(element_node, graph, parent_node, creation_info.document_namespace))
            except SPDXParsingError as err:
                logger.extend(err.get_messages())
        parsed_fields[element] = elements

    for triple, relationship_type in [
        ((None, SPDX_NAMESPACE.hasFile, None), RelationshipType.CONTAINS),
        ((None, SPDX_NAMESPACE.describesPackage, None), RelationshipType.DESCRIBES),
    ]:
        for parent_node, _, element_node in get_correctly_typed_triples(logger, graph, *triple):
            try:
                relationship = parse_implicit_relationship(
                    parent_node, relationship_type, element_node, graph, creation_info.document_namespace
                )
                if relationship not in parsed_fields["relationships"]:
                    parsed_fields["relationships"].append(relationship)

            except SPDXParsingError as err:
                logger.extend(err.get_messages())

    extracted_licensing_infos = []
    for _, _, extracted_licensing_info_node in get_correctly_typed_triples(
        logger, graph, None, SPDX_NAMESPACE.hasExtractedLicensingInfo
    ):
        try:
            extracted_licensing_infos.append(
                parse_extracted_licensing_info(extracted_licensing_info_node, graph, creation_info.document_namespace)
            )
        except SPDXParsingError as err:
            logger.extend(err.get_messages())
    parsed_fields["extracted_licensing_info"] = extracted_licensing_infos

    raise_parsing_error_if_logger_has_messages(logger)
    document = construct_or_raise_parsing_error(Document, parsed_fields)

    return document



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

import pytest

from spdx_tools.spdx3.model import ExternalMap
from tests.spdx3.fixtures import external_map_fixture, hash_fixture
from tests.spdx3.model.model_test_utils import get_property_names


def test_correct_initialization():
    external_map = external_map_fixture()

    for property_name in get_property_names(ExternalMap):
        assert getattr(external_map, property_name) is not None

    assert external_map.external_id == "https://spdx.test/tools-python/external_map_external_id"
    assert external_map.verified_using == [hash_fixture()]
    assert external_map.location_hint == "https://spdx.test/tools-python/external_map_location_hint"
    assert external_map.defining_document == "https://spdx.test/tools-python/defining_document"


def test_invalid_initialization():
    with pytest.raises(TypeError) as err:
        ExternalMap(234, None, ["location  hints"])

    assert len(err.value.args[0]) == 2
    for error in err.value.args[0]:
        assert error.startswith("SetterError ExternalMap:")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import Namespace

SPDX_NAMESPACE = Namespace("http://spdx.org/rdf/terms#")
POINTER_NAMESPACE = Namespace("http://www.w3.org/2009/pointers#")
REFERENCE_NAMESPACE = Namespace("http://spdx.org/rdf/references/")
LICENSE_NAMESPACE = Namespace("http://spdx.org/licenses/")



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import Type

from spdx_tools.spdx.jsonschema.checksum_properties import ChecksumProperty
from spdx_tools.spdx.jsonschema.converter import TypedConverter
from spdx_tools.spdx.jsonschema.json_property import JsonProperty
from spdx_tools.spdx.model import Checksum, ChecksumAlgorithm, Document


class ChecksumConverter(TypedConverter[Checksum]):
    def get_data_model_type(self) -> Type[Checksum]:
        return Checksum

    def get_json_type(self) -> Type[JsonProperty]:
        return ChecksumProperty

    def _get_property_value(
        self, checksum: Checksum, checksum_property: ChecksumProperty, _document: Document = None
    ) -> str:
        if checksum_property == ChecksumProperty.ALGORITHM:
            return algorithm_to_json_string(checksum.algorithm)
        elif checksum_property == ChecksumProperty.CHECKSUM_VALUE:
            return checksum.value


def algorithm_to_json_string(algorithm: ChecksumAlgorithm) -> str:
    name_with_dash: str = algorithm.name.replace("_", "-")
    if "BLAKE2B" in name_with_dash:
        return name_with_dash.replace("BLAKE2B", "BLAKE2b")
    return name_with_dash



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model.build import Build
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.element_writer import write_element_properties
from spdx_tools.spdx3.writer.console.hash_writer import write_hash
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_optional_heading


def write_build(build: Build, text_output: TextIO):
    text_output.write("## Build\n")
    write_element_properties(build, text_output)

    for property_name in Build.__annotations__.keys():
        if property_name == "config_source_digest":
            write_optional_heading(build.config_source_digest, "config_source_digest", text_output)
            for digest_hash in build.config_source_digest:
                write_hash(digest_hash, text_output, heading=False)
            continue

        write_value(property_name, getattr(build, property_name), text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import ExternalReference
from spdx_tools.spdx3.writer.console.console import write_value


def write_external_reference(external_reference: ExternalReference, text_output: TextIO):
    for property_name in ExternalReference.__annotations__.keys():
        write_value(property_name, getattr(external_reference, property_name), text_output)



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
import pytest

from spdx_tools.spdx3.model import Hash, HashAlgorithm
from tests.spdx3.fixtures import hash_fixture
from tests.spdx3.model.model_test_utils import get_property_names


def test_correct_initialization():
    hash = hash_fixture()

    for property_name in get_property_names(Hash):
        assert getattr(hash, property_name) is not None

    assert hash.algorithm == HashAlgorithm.SHA1
    assert hash.hash_value == "71c4025dd9897b364f3ebbb42c484ff43d00791c"
    assert hash.comment == "hashComment"


def test_invalid_initialization():
    with pytest.raises(TypeError) as err:
        Hash("SHA1", 345)

    assert len(err.value.args[0]) == 2
    for error in err.value.args[0]:
        assert error.startswith("SetterError Hash:")



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from beartype.typing import TextIO

from spdx_tools.spdx3.model import ExternalMap
from spdx_tools.spdx3.writer.console.console import write_value
from spdx_tools.spdx3.writer.console.hash_writer import write_hash
from spdx_tools.spdx.writer.tagvalue.tagvalue_writer_helper_functions import write_optional_heading


def write_external_map(external_map: ExternalMap, text_output: TextIO):
    write_value("external_id", external_map.external_id, text_output)
    write_optional_heading(external_map.verified_using, "verified using\n", text_output)
    for integrity_method in external_map.verified_using:
        # for now Hash is the only child class of the abstract class IntegrityMethod,
        # as soon as there are more inherited classes we need to implement a logic
        # that determines the correct write function for the "integrity_method" object
        write_hash(integrity_method, text_output, heading=False)
    write_value("location_hint", external_map.location_hint, text_output)



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0

from typing import List

import pytest

from spdx_tools.spdx.constants import DOCUMENT_SPDX_ID
from spdx_tools.spdx.model import Document, Relationship, RelationshipType, SpdxNoAssertion, SpdxNone
from spdx_tools.spdx.validation.relationship_validator import validate_relationship
from spdx_tools.spdx.validation.validation_message import SpdxElementType, ValidationContext, ValidationMessage
from tests.spdx.fixtures import document_fixture, relationship_fixture


@pytest.mark.parametrize("related_spdx_element", ["SPDXRef-Package", SpdxNoAssertion(), SpdxNone()])
def test_valid_relationship(related_spdx_element):
    relationship = Relationship(DOCUMENT_SPDX_ID, RelationshipType.DESCRIBES, related_spdx_element, comment="comment")
    validation_messages: List[ValidationMessage] = validate_relationship(relationship, "SPDX-2.3", document_fixture())

    assert validation_messages == []


@pytest.mark.parametrize(
    "spdx_element_id, related_spdx_element_id, expected_message",
    [
        (
            "SPDXRef-unknownFile",
            "SPDXRef-File",
            'did not find the referenced spdx_id "SPDXRef-unknownFile" in the SPDX document',
        ),
        (
            "SPDXRef-File",
            "SPDXRef-unknownFile",
            'did not find the referenced spdx_id "SPDXRef-unknownFile" in the SPDX document',
        ),
    ],
)
def test_unknown_spdx_id(spdx_element_id, related_spdx_element_id, expected_message):
    relationship: Relationship = relationship_fixture(
        spdx_element_id=spdx_element_id, related_spdx_element_id=related_spdx_element_id
    )
    validation_messages: List[ValidationMessage] = validate_relationship(relationship, "SPDX-2.3", document_fixture())

    expected = ValidationMessage(
        expected_message, ValidationContext(element_type=SpdxElementType.RELATIONSHIP, full_element=relationship)
    )

    assert validation_messages == [expected]


@pytest.mark.parametrize(
    "relationship, expected_message",
    [
        (
            Relationship(DOCUMENT_SPDX_ID, RelationshipType.SPECIFICATION_FOR, "SPDXRef-Package"),
            "RelationshipType.SPECIFICATION_FOR is not supported in SPDX-2.2",
        ),
        (
            Relationship(DOCUMENT_SPDX_ID, RelationshipType.REQUIREMENT_DESCRIPTION_FOR, "SPDXRef-Package"),
            "RelationshipType.REQUIREMENT_DESCRIPTION_FOR is not supported in SPDX-2.2",
        ),
    ],
)
def test_v2_3_only_types(relationship, expected_message):
    document: Document = document_fixture()

    validation_message: List[ValidationMessage] = validate_relationship(relationship, "SPDX-2.2", document)

    expected = [
        ValidationMessage(
            expected_message, ValidationContext(element_type=SpdxElementType.RELATIONSHIP, full_element=relationship)
        )
    ]

    assert validation_message == expected



# SPDX-FileCopyrightText: 2023 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from rdflib import RDFS, Graph, URIRef
from rdflib.term import Node

from spdx_tools.spdx.model import Relationship, RelationshipType
from spdx_tools.spdx.parser.logger import Logger
from spdx_tools.spdx.parser.parsing_functions import (
    construct_or_raise_parsing_error,
    raise_parsing_error_if_logger_has_messages,
)
from spdx_tools.spdx.parser.rdf.graph_parsing_functions import (
    parse_enum_value,
    parse_literal,
    parse_literal_or_no_assertion_or_none,
    parse_spdx_id,
)
from spdx_tools.spdx.rdfschema.namespace import SPDX_NAMESPACE


def parse_relationship(relationship_node: Node, graph: Graph, parent_node: URIRef, doc_namespace: str) -> Relationship:
    logger = Logger()
    spdx_element_id = parse_spdx_id(parent_node, doc_namespace, graph)

    relationship_type = parse_literal(
        logger,
        graph,
        relationship_node,
        SPDX_NAMESPACE.relationshipType,
        parsing_method=lambda x: parse_enum_value(x, RelationshipType, SPDX_NAMESPACE.relationshipType_),
    )
    related_spdx_element = parse_literal_or_no_assertion_or_none(
        logger,
        graph,
        relationship_node,
        SPDX_NAMESPACE.relatedSpdxElement,
        parsing_method=lambda x: parse_spdx_id(x, doc_namespace, graph),
    )

    comment = parse_literal(logger, graph, relationship_node, RDFS.comment)
    raise_parsing_error_if_logger_has_messages(logger, "Relationship")
    relationship = construct_or_raise_parsing_error(
        Relationship,
        dict(
            spdx_element_id=spdx_element_id,
            relationship_type=relationship_type,
            related_spdx_element_id=related_spdx_element,
            comment=comment,
        ),
    )

    return relationship


def parse_implicit_relationship(
    spdx_element_node: URIRef,
    relationship_type: RelationshipType,
    related_spdx_element_node: URIRef,
    graph: Graph,
    doc_namespace: str,
) -> Relationship:
    spdx_element_id = parse_spdx_id(spdx_element_node, doc_namespace, graph)
    related_spdx_element_id = parse_spdx_id(related_spdx_element_node, doc_namespace, graph)
    relationship = construct_or_raise_parsing_error(
        Relationship,
        dict(
            spdx_element_id=spdx_element_id,
            relationship_type=relationship_type,
            related_spdx_element_id=related_spdx_element_id,
        ),
    )
    return relationship



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class PackageVerificationCodeProperty(JsonProperty):
    PACKAGE_VERIFICATION_CODE_EXCLUDED_FILES = auto()
    PACKAGE_VERIFICATION_CODE_VALUE = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from enum import auto

from spdx_tools.spdx.jsonschema.json_property import JsonProperty


class RelationshipProperty(JsonProperty):
    SPDX_ELEMENT_ID = auto()
    COMMENT = auto()
    RELATED_SPDX_ELEMENT = auto()
    RELATIONSHIP_TYPE = auto()



# SPDX-FileCopyrightText: 2022 spdx contributors
#
# SPDX-License-Identifier: Apache-2.0
from datetime import datetime
from typing import Union
from unittest import mock
from unittest.mock import MagicMock, NonCallableMagicMock

import pytest

from spdx_tools.common.spdx_licensing import spdx_licensing
from spdx_tools.spdx.jsonschema.annotation_converter import AnnotationConverter
from spdx_tools.spdx.jsonschema.snippet_converter import SnippetConverter
from spdx_tools.spdx.jsonschema.snippet_properties import SnippetProperty
from spdx_tools.spdx.model import (
    Actor,
    ActorType,
    Annotation,
    AnnotationType,
    Document,
    Snippet,
    SpdxNoAssertion,
    SpdxNone,
)
from spdx_tools.spdx.model.spdx_no_assertion import SPDX_NO_ASSERTION_STRING
from spdx_tools.spdx.model.spdx_none import SPDX_NONE_STRING
from tests.spdx.fixtures import annotation_fixture, creation_info_fixture, document_fixture, snippet_fixture
from tests.spdx.mock_utils import assert_mock_method_called_with_arguments


@pytest.fixture
@mock.patch("spdx_tools.spdx.jsonschema.annotation_converter.AnnotationConverter", autospec=True)
def converter(annotation_converter_mock: MagicMock) -> SnippetConverter:
    converter = SnippetConverter()
    converter.annotation_converter = annotation_converter_mock()
    return converter


@pytest.mark.parametrize(
    "snippet_property,expected",
    [
        (SnippetProperty.SPDX_ID, "SPDXID"),
        (SnippetProperty.ANNOTATIONS, "annotations"),
        (SnippetProperty.ATTRIBUTION_TEXTS, "attributionTexts"),
        (SnippetProperty.COMMENT, "comment"),
        (SnippetProperty.COPYRIGHT_TEXT, "copyrightText"),
        (SnippetProperty.LICENSE_COMMENTS, "licenseComments"),
        (SnippetProperty.LICENSE_CONCLUDED, "licenseConcluded"),
        (SnippetProperty.LICENSE_INFO_IN_SNIPPETS, "licenseInfoInSnippets"),
        (SnippetProperty.NAME, "name"),
        (SnippetProperty.RANGES, "ranges"),
        (SnippetProperty.SNIPPET_FROM_FILE, "snippetFromFile"),
    ],
)
def test_json_property_names(converter: SnippetConverter, snippet_property: SnippetProperty, expected: str):
    assert converter.json_property_name(snippet_property) == expected


def test_json_type(converter: SnippetConverter):
    assert converter.get_json_type() == SnippetProperty


def test_data_model_type(converter: SnippetConverter):
    assert converter.get_data_model_type() == Snippet


def test_successful_conversion(converter: SnippetConverter):
    converter.annotation_converter.convert.return_value = "mock_converted_annotation"
    file_spdx_id = "fileSpdxId"
    snippet = Snippet(
        "spdxId",
        file_spdx_id=file_spdx_id,
        byte_range=(1, 2),
        line_range=(3, 4),
        license_concluded=spdx_licensing.parse("MIT and GPL-2.0"),
        license_info_in_snippet=[spdx_licensing.parse("MIT"), spdx_licensing.parse("GPL-2.0")],
        license_comment="licenseComment",
        copyright_text="copyrightText",
        comment="comment",
        name="name",
        attribution_texts=["attributionText1", "attributionText2"],
    )

    annotation = Annotation(
        snippet.spdx_id,
        AnnotationType.OTHER,
        Actor(ActorType.PERSON, "annotatorName"),
        datetime(2022, 12, 5),
        "other comment",
    )
    document = Document(creation_info_fixture(), snippets=[snippet], annotations=[annotation])
    converted_dict = converter.convert(snippet, document)

    assert converted_dict == {
        converter.json_property_name(SnippetProperty.SPDX_ID): "spdxId",
        converter.json_property_name(SnippetProperty.ANNOTATIONS): ["mock_converted_annotation"],
        converter.json_property_name(SnippetProperty.ATTRIBUTION_TEXTS): ["attributionText1", "attributionText2"],
        converter.json_property_name(SnippetProperty.COMMENT): "comment",
        converter.json_property_name(SnippetProperty.COPYRIGHT_TEXT): "copyrightText",
        converter.json_property_name(SnippetProperty.LICENSE_COMMENTS): "licenseComment",
        converter.json_property_name(SnippetProperty.LICENSE_CONCLUDED): "MIT AND GPL-2.0-only",
        converter.json_property_name(SnippetProperty.LICENSE_INFO_IN_SNIPPETS): ["MIT", "GPL-2.0-only"],
        converter.json_property_name(SnippetProperty.NAME): "name",
        converter.json_property_name(SnippetProperty.RANGES): [
            {
                "startPointer": {"reference": file_spdx_id, "offset": 1},
                "endPointer": {"reference": file_spdx_id, "offset": 2},
            },
            {
                "startPointer": {"reference": file_spdx_id, "lineNumber": 3},
                "endPointer": {"reference": file_spdx_id, "lineNumber": 4},
            },
        ],
        converter.json_property_name(SnippetProperty.SNIPPET_FROM_FILE): file_spdx_id,
    }


def test_null_values(converter: SnippetConverter):
    snippet = snippet_fixture(
        license_concluded=None,
        license_comment=None,
        copyright_text=None,
        comment=None,
        name=None,
        attribution_texts=[],
        license_info_in_snippet=[],
    )

    document = Document(creation_info_fixture(), snippets=[snippet])
    converted_dict = converter.convert(snippet, document)

    assert converter.json_property_name(SnippetProperty.LICENSE_CONCLUDED) not in converted_dict
    assert converter.json_property_name(SnippetProperty.LICENSE_COMMENTS) not in converted_dict
    assert converter.json_property_name(SnippetProperty.COPYRIGHT_TEXT) not in converted_dict
    assert converter.json_property_name(SnippetProperty.COMMENT) not in converted_dict
    assert converter.json_property_name(SnippetProperty.NAME) not in converted_dict
    assert converter.json_property_name(SnippetProperty.ANNOTATIONS) not in converted_dict
    assert converter.json_property_name(SnippetProperty.ATTRIBUTION_TEXTS) not in converted_dict
    assert converter.json_property_name(SnippetProperty.LICENSE_INFO_IN_SNIPPETS) not in converted_dict


def test_spdx_no_assertion(converter: SnippetConverter):
    snippet = snippet_fixture(license_concluded=SpdxNoAssertion(), license_info_in_snippet=[SpdxNoAssertion()])

    document = Document(creation_info_fixture(), snippets=[snippet])
    converted_dict = converter.convert(snippet, document)

    assert converted_dict[converter.json_property_name(SnippetProperty.LICENSE_CONCLUDED)] == SPDX_NO_ASSERTION_STRING
    assert converted_dict[converter.json_property_name(SnippetProperty.LICENSE_INFO_IN_SNIPPETS)] == [
        SPDX_NO_ASSERTION_STRING
    ]


def test_spdx_none(converter: SnippetConverter):
    snippet = snippet_fixture(license_concluded=SpdxNone(), license_info_in_snippet=[SpdxNone()])

    document = Document(creation_info_fixture(), snippets=[snippet])
    converted_dict = converter.convert(snippet, document)

    assert converted_dict[converter.json_property_name(SnippetProperty.LICENSE_CONCLUDED)] == SPDX_NONE_STRING
    assert converted_dict[converter.json_property_name(SnippetProperty.LICENSE_INFO_IN_SNIPPETS)] == [SPDX_NONE_STRING]


def test_snippet_annotations(converter: SnippetConverter):
    snippet = snippet_fixture(spdx_id="snippetId")
    document = document_fixture(snippets=[snippet])
    first_snippet_annotation = annotation_fixture(spdx_id=snippet.spdx_id)
    second_snippet_annotation = annotation_fixture(spdx_id=snippet.spdx_id)
    document_annotation = annotation_fixture(spdx_id=document.creation_info.spdx_id)
    package_annotation = annotation_fixture(spdx_id=document.packages[0].spdx_id)
    file_annotation = annotation_fixture(spdx_id=document.files[0].spdx_id)
    other_annotation = annotation_fixture(spdx_id="otherId")
    annotations = [
        first_snippet_annotation,
        second_snippet_annotation,
        document_annotation,
        package_annotation,
        file_annotation,
        other_annotation,
    ]
    document.annotations = annotations

    # Weird type hint to make warnings about unresolved references from the mock class disappear
    annotation_converter: Union[AnnotationConverter, NonCallableMagicMock] = converter.annotation_converter
    annotation_converter.convert.return_value = "mock_converted_annotation"

    converted_dict = converter.convert(snippet, document)

    assert_mock_method_called_with_arguments(
        annotation_converter, "convert", first_snippet_annotation, second_snippet_annotation
    )
    converted_file_annotations = converted_dict.get(converter.json_property_name(SnippetProperty.ANNOTATIONS))
    assert converted_file_annotations == ["mock_converted_annotation", "mock_converted_annotation"]
